# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import subprocess
import shutil
import gzip
from datetime import datetime
from typing import Optional

from spectre_core.config import get_spectre_data_dir_path, get_chunks_dir_path, TimeFormats

CALLISTO_INSTRUMENT_CODES = [
    "ALASKA-ANCHORAGE",
    "ALASKA-COHOE",
    "ALASKA-HAARP",
    "ALGERIA-CRAAG",
    "ALMATY",
    "AUSTRIA-Krumbach",
    "AUSTRIA-MICHELBACH",
    "AUSTRIA-OE3FLB",
    "AUSTRIA-UNIGRAZ",
    "Australia-ASSA",
    "BIR",
    "Croatia-Visnjan",
    "DENMARK",
    "EGYPT-Alexandria",
    "EGYPT-SpaceAgency",
    "FINLAND-Siuntio",
    "Finland-Kempele",
    "GERMANY-DLR",
    "GLASGOW",
    "GREENLAND",
    "HUMAIN",
    "HURBANOVO",
    "INDIA-GAURI",
    "INDIA-OOTY",
    "INDIA-UDAIPUR",
    "JAPAN-IBARAKI",
    "KASI",
    "MEXART",
    "MEXICO-FCFM-UANL",
    "MEXICO-LANCE-A",
    "MEXICO-LANCE-B",
    "MONGOLIA-UB",
    "MRO",
    "MRT3",
    "Malaysia-Banting",
    "NORWAY-EGERSUND",
    "NORWAY-NY-AALESUND",
    "NORWAY-RANDABERG",
    "POLAND-Grotniki",
    "ROMANIA",
    "ROSWELL-NM",
    "SPAIN-PERALEJOS",
    "SSRT",
    "SWISS-HB9SCT",
    "SWISS-HEITERSWIL",
    "SWISS-IRSOL",
    "SWISS-Landschlacht",
    "SWISS-MUHEN",
    "TRIEST",
    "TURKEY",
    "UNAM",
    "URUGUAY",
    "USA-BOSTON",
]

_temp_dir = os.path.join(get_spectre_data_dir_path(), "temp")


def _get_chunk_name(station: str, date: str, time: str, instrument_code: str) -> str:
    dt = datetime.strptime(f"{date}T{time}", '%Y%m%dT%H%M%S')
    formatted_time = dt.strftime(TimeFormats.DATETIME)
    return f"{formatted_time}_callisto-{station.lower()}-{instrument_code}.fits"


def _get_chunk_components(gz_path: str):
    file_name = os.path.basename(gz_path)
    if not file_name.endswith(".fit.gz"):
        raise ValueError(f"Unexpected file extension in {file_name}. Expected .fit.gz")
    
    file_base_name = file_name.rstrip(".fit.gz")
    parts = file_base_name.split('_')
    if len(parts) != 4:
        raise ValueError("Filename does not conform to the expected format of [station]_[date]_[time]_[instrument_code]")
    
    return parts


def _get_chunk_path(gz_path: str) -> str:
    station, date, time, instrument_code = _get_chunk_components(gz_path)
    fits_chunk_name = _get_chunk_name(station, date, time, instrument_code)
    chunk_start_time = fits_chunk_name.split('_')[0]
    chunk_start_datetime = datetime.strptime(chunk_start_time, TimeFormats.DATETIME)
    chunk_parent_path = get_chunks_dir_path(year = chunk_start_datetime.year,
                                            month = chunk_start_datetime.month,
                                            day = chunk_start_datetime.day)
    if not os.path.exists(chunk_parent_path):
        os.makedirs(chunk_parent_path)
    return os.path.join(chunk_parent_path, fits_chunk_name)


def _unzip_file_to_chunks(gz_path: str):
    fits_path = _get_chunk_path(gz_path)
    with gzip.open(gz_path, 'rb') as f_in, open(fits_path, 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)


def _unzip_to_chunks():
    for entry in os.scandir(_temp_dir):
        if entry.is_file() and entry.name.endswith('.gz'):
            _unzip_file_to_chunks(entry.path)
            os.remove(entry.path)


def _wget_callisto_data(instrument_code: str, 
                       year: int, 
                       month: int, 
                       day: int):
    date_str = f"{year:04d}/{month:02d}/{day:02d}"
    base_url = f"http://soleil.i4ds.ch/solarradio/data/2002-20yy_Callisto/{date_str}/"
    command = [
        'wget', '-r', '-l1', '-nd', '-np', 
        '-R', '.tmp',
        '-A', f'{instrument_code}*.fit.gz',
        '-P', _temp_dir,
        base_url
    ]

    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        print(f"An error occurred: {e}")


def download_callisto_data(instrument_code: Optional[str], 
                           year: Optional[int], 
                           month: Optional[int], 
                           day: Optional[int]):


    if (year is None) or (month is None) or (day is None):
        raise ValueError(f"All of year, month and day should be specified")
    
    if not os.path.exists(_temp_dir):
        os.mkdir(_temp_dir)

    if instrument_code not in CALLISTO_INSTRUMENT_CODES:
        raise ValueError(f"No match found for '{instrument_code}'. Expected one of {CALLISTO_INSTRUMENT_CODES}")

    _wget_callisto_data(instrument_code, year, month, day)
    _unzip_to_chunks()
    shutil.rmtree(_temp_dir)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later


from logging import getLogger
_LOGGER = getLogger(__name__)

import os
import logging
from dataclasses import dataclass
from typing import Optional
import warnings
from collections import OrderedDict
from datetime import datetime

from spectre_core._file_io import TextHandler
from spectre_core.config import get_logs_dir_path, TimeFormats

PROCESS_TYPES = [
    "user", 
    "worker"
]

def _validate_process_type(process_type: str
) -> None:
    if process_type not in PROCESS_TYPES:
        raise ValueError(f"Invalid process type: {process_type}. Expected one of {PROCESS_TYPES}")


class LogHandler(TextHandler):
    def __init__(self, 
                 datetime_stamp: str, 
                 pid: str, 
                 process_type: str):
        self._datetime_stamp = datetime_stamp
        self._pid = pid
        _validate_process_type(process_type)
        self._process_type = process_type

        dt = datetime.strptime(datetime_stamp, TimeFormats.DATETIME)
        parent_path = get_logs_dir_path(dt.year, dt.month, dt.day)
        base_file_name = f"{datetime_stamp}_{pid}_{process_type}"

        super().__init__(parent_path, base_file_name, extension = "log")
    

    @property
    def datetime_stamp(self) -> str:
        return self._datetime_stamp
    

    @property
    def pid(self) -> str:
        return self._pid
    

    @property
    def process_type(self) -> str:
        return self._process_type


class LogHandlers:
    def __init__(self, 
                 process_type: Optional[str] = None, 
                 year: Optional[int] = None, 
                 month: Optional[int] = None, 
                 day: Optional[int] = None):
        self._log_handler_map: dict[str, LogHandler] = OrderedDict()
        self._process_type = process_type
        self.set_date(year, month, day)


    @property
    def process_type(self) -> str:
        return self._process_type
    

    @property
    def year(self) -> Optional[int]:
        return self._year


    @property 
    def month(self) -> Optional[int]:
        return self._month
    

    @property
    def day(self) -> Optional[int]:
        return self._day


    @property
    def logs_dir_path(self) -> str:
        return get_logs_dir_path(self.year, self.month, self.day)


    @property 
    def log_handler_map(self) -> dict[str, LogHandler]:
        if not self._log_handler_map: # check for empty dictionary
            self._update_log_handler_map()
        return self._log_handler_map
        

    @property
    def log_handler_list(self) -> list[LogHandler]:
        return list(self.log_handler_map.values())


    @property
    def num_logs(self) -> int:
        return len(self.log_handler_list) 


    @property
    def file_names(self) -> list[str]:
        return list(self.log_handler_map.keys())


    def set_date(self, 
                 year: Optional[int],
                 month: Optional[int],
                 day: Optional[int]) -> None:
        self._year = year
        self._month = month
        self._day = day
        self._update_log_handler_map()


    def _update_log_handler_map(self) -> None:
        log_files = [f for (_, _, files) in os.walk(self.logs_dir_path) for f in files]

        if not log_files:
            warning_message = "No logs found, setting log map to an empty dictionary"
            _LOGGER.warning(warning_message)
            warnings.warn(warning_message)
            return

        for log_file in log_files:
            file_name, _ = os.path.splitext(log_file)
            log_start_time, pid, process_type = file_name.split("_")

            if self.process_type and process_type != self.process_type:
                continue

            self._log_handler_map[file_name] = LogHandler(log_start_time, pid, process_type)

        self._log_handler_map = OrderedDict(sorted(self._log_handler_map.items()))


    def update(self) -> None:
        """Public alias for setting log handler map"""
        self._update_log_handler_map()


    def __iter__(self):
        yield from self.log_handler_list


    def get_from_file_name(self, 
                           file_name: str) -> LogHandler:
        # auto strip the extension if present
        file_name, _ = os.path.splitext(file_name)
        try:
            return self.log_handler_map[file_name]
        except KeyError:
            raise FileNotFoundError(f"Log handler for file name '{file_name}' not found in log map")


    def get_from_pid(self, 
                     pid: str) -> LogHandler:
        for log_handler in self.log_handler_list:
            if log_handler.pid == pid:
                return log_handler
        raise FileNotFoundError(f"Log handler for PID '{pid}' not found in log map")


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from copy import deepcopy
from typing import Any
from dataclasses import dataclass

from ._parameters import Parameter, Parameters
from ._pconstraints import PConstraint
from ._ptemplates import (
    PTemplate, 
    PNames,
    get_base_ptemplate
)

class CaptureTemplate:
    """A managed collection of PTemplates"""
    def __init__(self):
        self._ptemplates: dict[str, PTemplate] = {}


    @property
    def name_list(self) -> list[str]:
        """List the names of all stored PTemplates."""
        return list(self._ptemplates.keys())
    

    def add_ptemplate(self,
                      ptemplate: PTemplate) -> None:
        """Add a ptemplate to this capture template."""
        self._ptemplates[ptemplate.name] = ptemplate


    def get_ptemplate(self,
                      parameter_name: str) -> PTemplate:
        """Get the ptemplate corresponding with the parameter name."""
        if parameter_name not in self._ptemplates:
            raise ValueError(f"Parameter with name '{parameter_name}' is not found in the template. "
                             f"Expected one of {self.name_list}")   
        return self._ptemplates[parameter_name]
      

    def __apply_parameter_template(self,
                                   parameter: Parameter):
        """Apply the corresponding parameter template to the input parameter"""
        ptemplate = self.get_ptemplate(parameter.name)
        parameter.value = ptemplate.apply_template(parameter.value)


    def __apply_parameter_templates(self,
                                    parameters: Parameters) -> None:
        """Apply the corresponding parameter template to all explictly specified parameters"""
        for parameter in parameters:
            self.__apply_parameter_template(parameter)

    
    def __fill_missing_with_defaults(self,
                                     parameters: Parameters) -> None:
        """For any missing parameters (as per the capture template), use the corresponding default value."""
        for ptemplate in self:
            if ptemplate.name not in parameters.name_list:
                parameter = ptemplate.make_parameter()
                parameters.add_parameter(parameter.name, 
                                         parameter.value)


    def apply_template(self,
                       parameters: Parameters) -> Parameters:
        """Validate parameters, fill missing with defaults, and return anew."""
        self.__apply_parameter_templates(parameters)
        self.__fill_missing_with_defaults(parameters)
        return parameters


    def __iter__(self):
        """Iterate over stored ptemplates"""
        yield from self._ptemplates.values() 


    def set_default(self, 
                    parameter_name: str, 
                    default: Any) -> None:
        """Set the default of an existing ptemplate."""
        self.get_ptemplate(parameter_name).default = default


    def set_defaults(self, 
                     *ptuples: tuple[str, Any]) -> None:
        """Update defaults for multiple ptemplates."""
        for (parameter_name, default) in ptuples:
            self.set_default(parameter_name, default)


    def enforce_default(self,
                        parameter_name: str) -> None:
        """Enforce the default of an existing ptemplate"""
        self.get_ptemplate(parameter_name).enforce_default = True


    def enforce_defaults(self, 
                         *parameter_names: str) -> None:
        """Enforce defaults for multiple parameter names."""
        for name in parameter_names:
            self.enforce_default(name)


    def add_pconstraint(self,
                        parameter_name: str,
                        pconstraints: list[PConstraint]) -> None:
        """Add a pconstraint to an existing ptemplate"""
        for pconstraint in pconstraints:
            self.get_ptemplate(parameter_name).add_pconstraint(pconstraint)


    def to_dict(self) -> dict:
        return {ptemplate.name: ptemplate.to_dict() for ptemplate in self}


def make_base_capture_template(*parameter_names: str):
    """Make a capture template, composed entirely of base ptemplates."""
    capture_template = CaptureTemplate()
    for name in parameter_names:
        capture_template.add_ptemplate( get_base_ptemplate(name) )
    return capture_template


@dataclass(frozen=True)
class CaptureModes:
    """Pre-defined capture types"""
    FIXED_CENTER_FREQUENCY: str = "fixed-center-frequency"
    SWEPT_CENTER_FREQUENCY: str = "swept-center-frequency"


def _make_fixed_frequency_capture_template(
) -> CaptureTemplate:
    """The absolute minimum required parameters for any fixed frequency capture template."""
    capture_template = make_base_capture_template(
        PNames.BATCH_SIZE,
        PNames.CENTER_FREQUENCY,
        PNames.CHUNK_KEY,
        PNames.EVENT_HANDLER_KEY,
        PNames.FREQUENCY_RESOLUTION,
        PNames.INSTRUMENT,
        PNames.OBS_ALT,
        PNames.OBS_LAT,
        PNames.OBS_LON,
        PNames.OBJECT,
        PNames.ORIGIN,
        PNames.SAMPLE_RATE,
        PNames.TELESCOPE,
        PNames.TIME_RANGE,
        PNames.TIME_RESOLUTION,
        PNames.WATCH_EXTENSION,
        PNames.WINDOW_HOP,
        PNames.WINDOW_SIZE,
        PNames.WINDOW_TYPE,
    )
    capture_template.set_defaults(
            (PNames.EVENT_HANDLER_KEY,     CaptureModes.FIXED_CENTER_FREQUENCY),
            (PNames.CHUNK_KEY,             CaptureModes.FIXED_CENTER_FREQUENCY),
            (PNames.WATCH_EXTENSION,       "bin")
    )
    capture_template.enforce_defaults(
        PNames.EVENT_HANDLER_KEY,
        PNames.CHUNK_KEY,
        PNames.WATCH_EXTENSION
    )
    return capture_template

def _make_swept_frequency_capture_template(
) -> CaptureTemplate:
    """The absolute minimum required parameters for any swept frequency capture template."""
    capture_template = make_base_capture_template(
        PNames.BATCH_SIZE,
        PNames.CHUNK_KEY,
        PNames.EVENT_HANDLER_KEY,
        PNames.FREQUENCY_RESOLUTION,
        PNames.FREQUENCY_STEP,
        PNames.INSTRUMENT,
        PNames.MAX_FREQUENCY,
        PNames.MIN_FREQUENCY,
        PNames.OBS_ALT,
        PNames.OBS_LAT,
        PNames.OBS_LON,
        PNames.OBJECT,
        PNames.ORIGIN,
        PNames.SAMPLE_RATE,
        PNames.SAMPLES_PER_STEP,
        PNames.TELESCOPE,
        PNames.TIME_RANGE,
        PNames.TIME_RESOLUTION,
        PNames.WATCH_EXTENSION,
        PNames.WINDOW_HOP,
        PNames.WINDOW_SIZE,
        PNames.WINDOW_TYPE)
    capture_template.set_defaults(
            (PNames.EVENT_HANDLER_KEY,     CaptureModes.SWEPT_CENTER_FREQUENCY),
            (PNames.CHUNK_KEY,             CaptureModes.SWEPT_CENTER_FREQUENCY),
            (PNames.WATCH_EXTENSION,       "bin")
    )
    capture_template.enforce_defaults(
        PNames.EVENT_HANDLER_KEY,
        PNames.CHUNK_KEY,
        PNames.WATCH_EXTENSION
    )
    return capture_template


_base_capture_templates = {
    CaptureModes.FIXED_CENTER_FREQUENCY: _make_fixed_frequency_capture_template(),
    CaptureModes.SWEPT_CENTER_FREQUENCY: _make_swept_frequency_capture_template()
}

def get_base_capture_template(
       capture_mode: str
) -> CaptureTemplate:
    """Create a fresh deep copy of a pre-defined capture template"""
    if capture_mode not in _base_capture_templates:
        raise KeyError(f"No capture template found for the capture mode '{capture_mode}'. "
                       f"Expected one of {list(_base_capture_templates.keys())}")
    return deepcopy( _base_capture_templates[capture_mode] )


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from spectre_core.capture_configs import CaptureConfig, PNames
from spectre_core.exceptions import ChunkNotFoundError
from ._register import chunk_map
from ._base import BaseChunk

def _get_chunk(chunk_key: str) -> BaseChunk:
    Chunk = chunk_map.get(chunk_key)
    if Chunk is None:
        valid_chunk_keys = list(chunk_map.keys())
        raise ChunkNotFoundError(f"No chunk found for the chunk key: {chunk_key}. Valid chunk keys are: {valid_chunk_keys}")
    return Chunk

def get_chunk_from_tag(tag: str) -> BaseChunk:
    # if we are dealing with a callisto chunk, the chunk key is equal to the tag
    if "callisto" in tag:
        chunk_key = "callisto"
    # otherwise, we fetch the chunk key from the capture config
    else:
        capture_config= CaptureConfig(tag)
        chunk_key = capture_config.get_parameter_value(PNames.CHUNK_KEY)
    return _get_chunk(chunk_key)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Tuple
from dataclasses import dataclass

import numpy as np

from spectre_core.exceptions import InvalidSweepMetadataError
from spectre_core.capture_configs import CaptureModes
from ._fixed_center_frequency import BinChunk, FitsChunk
from .._register import register_chunk
from .._base import BaseChunk, ChunkFile


@dataclass
class SweepMetadata:
    """Wrapper for metadata required to assign center frequencies to each IQ sample in the chunk.
    
    center_frequencies is an ordered list containing all the center frequencies that the IQ samples
    were collected at. Typically, these will be ordered in "steps", where each step corresponds to
    IQ samples collected at a constant center frequency:

    (freq_0, freq_1, ..., freq_M, freq_0, freq_1, ..., freq_M, ...), freq_0 < freq_1 < ... < freq_M

    The n'th element of the num_samples list, tells us how many samples were collected at the n'th
    element of center_frequencies:
chunks.library.fixed_center_frequency.chunk import (
    BinChunk, FitsChunk
)
    Number of samples: (num_samples_at_freq_0, num_samples_at_freq_1, ...)

    Both these lists together allow us to map for each IQ sample, the center frequency it was collected at.
    """
    center_frequencies: np.ndarray
    num_samples: np.ndarray


@register_chunk(CaptureModes.SWEPT_CENTER_FREQUENCY)
class _Chunk(BaseChunk):
    def __init__(self, chunk_start_time, tag):
        super().__init__(chunk_start_time, tag)

        self.add_file(BinChunk(self.chunk_parent_path, self.chunk_name))
        self.add_file(FitsChunk(self.chunk_parent_path, self.chunk_name))
        self.add_file(HdrChunk(self.chunk_parent_path, self.chunk_name))

    
class HdrChunk(ChunkFile):
    def __init__(self, chunk_parent_path: str, chunk_name: str):
        super().__init__(chunk_parent_path, chunk_name, "hdr")

    def read(self) -> Tuple[int, SweepMetadata]:
        hdr_contents = self._read_file_contents()
        millisecond_correction = self._get_millisecond_correction(hdr_contents)
        center_frequencies = self._get_center_frequencies(hdr_contents)
        num_samples = self._get_num_samples(hdr_contents)
        self._validate_frequencies_and_samples(center_frequencies, 
                                               num_samples)
        return millisecond_correction, SweepMetadata(center_frequencies, num_samples)
        

    def _read_file_contents(self) -> np.ndarray:
        with open(self.file_path, "rb") as fh:
            return np.fromfile(fh, dtype=np.float32)


    def _get_millisecond_correction(self, hdr_contents: np.ndarray) -> int:
        ''' Millisecond correction is an integral quantity, but stored in the detached header as a 32-bit float.'''
        millisecond_correction_as_float = float(hdr_contents[0])

        if not millisecond_correction_as_float.is_integer():
            raise TypeError(f"Expected integer value for millisecond correction, but got {millisecond_correction_as_float}")
        
        return int(millisecond_correction_as_float)


    def _get_center_frequencies(self, hdr_contents: np.ndarray) -> np.ndarray:
        ''' 
        Detached header contents are stored in (center_freq_i, num_samples_at_center_freq_i) pairs
        Return only a list of center frequencies, by skipping over file contents in twos.
        '''
        return hdr_contents[1::2]


    def _get_num_samples(self, hdr_contents: np.ndarray) -> np.ndarray:
        ''' 
        Detached header contents are stored in (center_freq_i, num_samples_at_center_freq_i) pairs
        Return only the number of samples at each center frequency, by skipping over file contents in twos.
        Number of samples is an integral quantity, but stored in the detached header as a 32-bit float.
        Types are checked before return.
        '''
        num_samples_as_float = hdr_contents[2::2]
        if not all(num_samples_as_float == num_samples_as_float.astype(int)):
            raise InvalidSweepMetadataError("Number of samples per frequency is expected to describe an integer")
        return num_samples_as_float.astype(int)


    def _validate_frequencies_and_samples(self, center_frequencies: np.ndarray, num_samples: np.ndarray) -> None:
        """Validates that the center frequencies and the number of samples arrays have the same length."""
        if len(center_frequencies) != len(num_samples):
            raise InvalidSweepMetadataError("Center frequencies and number of samples arrays are not the same length")



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
from typing import Optional, Tuple
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes

from spectre_core.spectrograms import TimeTypes
from ._base import BasePanel
from ._format import PanelFormat, DEFAULT_FORMAT
from ._panels import PanelNames


def _is_cuts_panel(panel: BasePanel) -> bool:
    return (panel.name == PanelNames.FREQUENCY_CUTS or panel.name == PanelNames.TIME_CUTS)


def _is_spectrogram_panel(panel: BasePanel) -> bool:
    return (panel.name == PanelNames.SPECTROGRAM)


class PanelStack:
    def __init__(self, 
                 panel_format: PanelFormat = DEFAULT_FORMAT,
                 time_type: str = TimeTypes.SECONDS,
                 figsize: Tuple[int, int] = (10, 10)):
        self._panel_format = panel_format
        self._time_type = time_type
        self._figsize = figsize

        self._panels: list[BasePanel] = []
        self._superimposed_panels: list[BasePanel] = []
        self._fig: Optional[Figure] = None
        self._axs: Optional[np.ndarray[Axes]] = None


    @property
    def time_type(self) -> str:
        return self._time_type


    @property
    def panels(self) -> list[BasePanel]:
        return sorted(self._panels, key=lambda panel: panel.x_axis_type)


    @property
    def fig(self) -> Optional[Figure]:
        return self._fig


    @property
    def axs(self) -> Optional[np.ndarray[Axes]]:
        return np.atleast_1d(self._axs)


    @property
    def num_panels(self) -> int:
        return len(self._panels)


    def add_panel(self, 
                  panel: BasePanel,
                  identifier: Optional[str] = None) -> None:
        panel.panel_format = self._panel_format
        panel.time_type    = self._time_type
        if identifier: 
            panel.identifier = identifier
        self._panels.append(panel)


    def superimpose_panel(self, 
                          panel: BasePanel,
                          identifier: Optional[str] = None) -> None:
        if identifier:
            panel.identifier = identifier
        panel.panel_format = self._panel_format
        self._superimposed_panels.append(panel)


    def _init_plot_style(self) -> None:
        plt.style.use(self._panel_format.style)

        plt.rc('font'  , size=self._panel_format.small_size)
        plt.rc('axes'  , titlesize=self._panel_format.medium_size, 
                         labelsize=self._panel_format.medium_size)
        plt.rc('xtick' , labelsize=self._panel_format.small_size)
        plt.rc('ytick' , labelsize=self._panel_format.small_size)
        plt.rc('legend', fontsize=self._panel_format.small_size)
        plt.rc('figure', titlesize=self._panel_format.large_size)


    def _create_figure_and_axes(self) -> None:
        self._fig, self._axs = plt.subplots(self.num_panels, 
                                            1, 
                                            figsize=self._figsize, 
                                            layout="constrained")


    def _assign_axes(self) -> None:
        shared_axes = {}
        for i, panel in enumerate(self.panels):
            panel.ax = self.axs[i]
            panel.fig = self._fig
            if panel.x_axis_type in shared_axes:
                panel.ax.sharex(shared_axes[panel.x_axis_type])
            else:
                shared_axes[panel.x_axis_type] = panel.ax


    def _overlay_cuts(self, cuts_panel: BasePanel) -> None:
        """Given a cuts panel, finds any corresponding spectrogram panels and adds the appropriate overlay"""
        for panel in self.panels:
            is_corresponding_panel = _is_spectrogram_panel(panel) and (panel.tag == cuts_panel.tag)
            if is_corresponding_panel:
                panel.overlay_cuts(cuts_panel)


    def _overlay_superimposed_panels(self) -> None:
        for super_panel in self._superimposed_panels:
            for panel in self._panels:
                if panel.name == super_panel.name and (panel.identifier == super_panel.identifier):
                    super_panel.ax, super_panel.fig = panel.ax, self._fig
                    super_panel.draw()
                    if _is_cuts_panel(super_panel):
                        self._overlay_cuts(super_panel)


    def show(self) -> None:
        self._init_plot_style()
        self._create_figure_and_axes()
        self._assign_axes()
        last_panel_per_axis = {panel.x_axis_type: panel for panel in self.panels}
        for panel in self.panels:
            panel.draw()
            panel.annotate_y_axis()
            if panel == last_panel_per_axis[panel.x_axis_type]:
                panel.annotate_x_axis()
            else:
                panel.hide_x_axis_labels()
            if _is_cuts_panel(panel):
                self._overlay_cuts(panel)
        self._overlay_superimposed_panels()
        plt.show()



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass

from spectre_core._file_io import JsonHandler
from spectre_core.config import get_configs_dir_path
from spectre_core.exceptions import InvalidTagError
from ._parameters import (
    Parameter, 
    Parameters,
    make_parameters
)

@dataclass
class _CaptureConfigKeys:
    RECEIVER_NAME = "receiver_name"
    RECEIVER_MODE = "receiver_mode"
    PARAMETERS    = "parameters"


class CaptureConfig(JsonHandler):
    def __init__(self,
                 tag: str):
        self._validate_tag(tag)
        self._tag = tag
        super().__init__(get_configs_dir_path(),
                         f"capture_{tag}")
        
    @property
    def tag(self) -> str:
        """Unique identifier for the capture config."""
        return self._tag


    def _validate_tag(self, 
                      tag: str) -> None:
        if "_" in tag:
            raise InvalidTagError(f"Tags cannot contain an underscore. Received {tag}")
        if "callisto" in tag:
            raise InvalidTagError(f'"callisto" cannot be a substring in a native tag. Received "{tag}"')
    

    @property
    def receiver_name(self) -> str:
        """The name of the receiver which created the capture config."""
        return self.dict[_CaptureConfigKeys.RECEIVER_NAME]
    

    @property
    def receiver_mode(self) -> str:
        """The mode of the receiver which created the capture config."""
        return self.dict[_CaptureConfigKeys.RECEIVER_MODE]
    

    @property
    def parameters(self) -> Parameters:
        """The parameters stored inside the capture config."""
        return make_parameters( self.dict[_CaptureConfigKeys.PARAMETERS] )


    def get_parameter(self, 
                      name: str) -> Parameter:
        return self.parameters.get_parameter(name)
    

    def get_parameter_value(self,
                            name: str) -> Parameter:
        return self.parameters.get_parameter_value(name)


    def save_parameters(self,
                        receiver_name: str,
                        receiver_mode: str,
                        parameters: Parameters,
                        force: bool = False):
        """Write the input parameters to file."""
        d = {
            _CaptureConfigKeys.RECEIVER_MODE: receiver_mode,
            _CaptureConfigKeys.RECEIVER_NAME: receiver_name,
            _CaptureConfigKeys.PARAMETERS   : parameters.to_dict()
        }
        self.save(d,
                  force=force)


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from ._callisto import download_callisto_data, CALLISTO_INSTRUMENT_CODES

__all__ = [
    "download_callisto_data", "CALLISTO_INSTRUMENT_CODES"
]


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, Optional, TypeVar

T = TypeVar('T')

class Parameter:
    """A named value."""
    def __init__(self, 
                 name: str,
                 value: Optional[T] = None):
        self._name = name
        self._value: Optional[T] = value


    @property
    def name(self) -> str:
        """The parameter name."""
        return self._name
    

    @property
    def value(self) -> Optional[T]:
        """The parameter value."""
        return self._value
    
    
    @value.setter
    def value(self, v: Optional[T]) -> None:
        """Update the parameter value."""
        self._value = v


class Parameters:
    """A collection of parameters."""
    def __init__(self):
        self._parameters: dict[str, Parameter] = {}
    

    @property
    def name_list(self) -> list[str]:
        """List the names of stored parameters."""
        return list(self._parameters.keys())


    def add_parameter(self, 
                      name: str,
                      value: Optional[T] = None,
                      force: bool = False) -> None:
        """Add a new parameter."""
        if name in self._parameters and not force:
            raise ValueError(f"Cannot add a parameter with name '{name}', "
                             f"since a parameter already exists with that name. "
                             f"You can overrride this functionality with 'force', "
                             f"to overwrite the existing parameter.")
        self._parameters[name] = Parameter(name, value)


    def get_parameter(self, 
                      name: str) -> Parameter:
        """Get the parameter with the corresponding name."""
        if name not in self._parameters:
            raise KeyError(f"Parameter with name '{name}' does not exist. "
                           f"Expected one of {self.name_list}")      
        return self._parameters[name]


    def get_parameter_value(self,
                            name: str) -> Optional[T]:
        """Get the value of the parameter with the corresponding name."""
        return self.get_parameter(name).value
    

    def __iter__(self):
        """Iterate over stored parameters"""
        yield from self._parameters.values() 

    
    def to_dict(self) -> dict:
        """Convert the class instance to an equivalent dictionary representation"""
        return {p.name: p.value for p in self}
    
def _parse_string_parameter(string_parameter: str) -> list[str]:
    """Parse string of the form 'a=b; into a list of the form [a, b]"""
    if not string_parameter or '=' not in string_parameter:
        raise ValueError(f"Invalid format: '{string_parameter}'. Expected 'KEY=VALUE'.")
    if string_parameter.startswith('=') or string_parameter.endswith('='):
        raise ValueError(f"Invalid format: '{string_parameter}'. Expected 'KEY=VALUE'.")
    # remove leading and trailing whitespace.
    string_parameter = string_parameter.strip()
    return string_parameter.split('=', 1)
    

def parse_string_parameters(string_parameters: list[str]) -> dict[str, str]:
    """Parses a list of strings of the form 'a=b'; into a dictionary mapping each 'a' to each 'b'"""
    d = {}
    for string_parameter in string_parameters:
        k, v = _parse_string_parameter(string_parameter)
        d[k] = v
    return d
    

def make_parameters(d: dict[str, Any]):
    """Make an instance of parameters based on the input dictionary"""
    parameters = Parameters()
    for k, v in d.items():
        parameters.add_parameter(k, v)
    return parameters


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Optional

from spectre_core.exceptions import ReceiverNotFoundError
from ._register import receivers
from ._base import BaseReceiver

# used to fetch an instance of the receiver class
def get_receiver(receiver_name: str, mode: Optional[str] = None) -> BaseReceiver:
    Receiver = receivers.get(receiver_name)
    if Receiver is None:
        valid_receivers = list(receivers.keys())
        raise ReceiverNotFoundError(f"No class found for the receiver: {receiver_name}. "
                                    f"Please specify one of the following receivers {valid_receivers}")
    return Receiver(receiver_name, 
                    mode = mode)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import json
from abc import ABC, abstractmethod
from typing import Any, Optional


class BaseFileHandler(ABC):
    def __init__(self, 
                 parent_path: str, 
                 base_file_name: str, 
                 extension: Optional[str] = None):
        self._parent_path = parent_path
        self._base_file_name = base_file_name
        self._extension = extension

        
    @abstractmethod
    def read(self) -> Any:
        pass
 

    @property
    def parent_path(self) -> str:
        return self._parent_path
    

    @property
    def base_file_name(self) -> str:
        return self._base_file_name
    

    @property
    def extension(self) -> Optional[str]:
        return self._extension
    

    @property
    def file_name(self) -> str:
        return self._base_file_name if (self._extension is None) else f"{self._base_file_name}.{self._extension}"
    

    @property
    def file_path(self) -> str:
        return os.path.join(self._parent_path, self.file_name)
    
    
    @property
    def exists(self) -> bool:
        return os.path.exists(self.file_path) 


    def make_parent_path(self) -> None:
        os.makedirs(self.parent_path, exist_ok=True) 
    

    def delete(self,
               ignore_if_missing: bool = False) -> None:
        if not self.exists and not ignore_if_missing:
            raise FileNotFoundError(f"{self.file_name} does not exist, and so cannot be deleted")
        else:
            os.remove(self.file_path)
    

    def cat(self) -> None:
        print(self.read())


class JsonHandler(BaseFileHandler):
    def __init__(self, 
                 parent_path: str, 
                 base_file_name: str,
                 extension: str = "json",
                 **kwargs):
        
        self._dict = None # cache
        super().__init__(parent_path, 
                         base_file_name, 
                         extension,
                         **kwargs)
    
    
    def read(self) -> dict[str, Any]:
        with open(self.file_path, 'r') as f:
            return json.load(f)
        

    def save(self, 
             d: dict, 
             force: bool = False) -> None:
        self.make_parent_path()

        if self.exists:
            if force:
                pass
            else:
                raise RuntimeError((f"{self.file_name} already exists, write has been abandoned. "
                                    f"You can override this functionality with `force`"))

        with open(self.file_path, 'w') as file:
                json.dump(d, file, indent=4)


    @property
    def dict(self) -> dict[str, Any]:
        if self._dict is None:
            self._dict = self.read()
        return self._dict
    

class TextHandler(BaseFileHandler):
    def __init__(self, 
                 parent_path: str,
                 base_file_name: str, 
                 extension: str = "txt",
                 **kwargs):
        super().__init__(parent_path,
                         base_file_name,
                         extension, 
                         **kwargs)
    

    def read(self) -> dict:
        with open(self.file_path, 'r') as f:
            return f.read()


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
import signal

from gnuradio import gr
from gnuradio import spectre

from spectre_core.capture_configs import Parameters, PNames
from spectre_core.config import get_chunks_dir_path

from spectre_core.capture_configs import Parameters


def capture(tag: str,
            parameters: Parameters, 
            top_block_cls: gr.top_block,
            max_noutput_items: int = 10000000):
    tb: gr.top_block = top_block_cls(tag,
                                     parameters)

    def sig_handler(sig=None, frame=None):
        tb.stop()
        tb.wait()
        sys.exit(0)

    signal.signal(signal.SIGINT, sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)

    tb.start(max_noutput_items)
    tb.wait()


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from ._format import PanelFormat
from ._panels import Panels
from ._panel_stack import PanelStack

__all__ = [
    "PanelFormat", "Panels", "PanelStack"
]


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from ._paths import (
    get_spectre_data_dir_path, get_chunks_dir_path, get_configs_dir_path, get_logs_dir_path
)
from ._time_formats import (
    TimeFormats
)

__all__ = [
    "get_spectre_data_dir_path",
    "get_chunks_dir_path",
    "get_configs_dir_path",
    "get_logs_dir_path",
    "DEFAULT_DATE_FORMAT",
    "DEFAULT_TIME_FORMAT",
    "DEFAULT_DATETIME_FORMAT"
]


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime, timedelta
from typing import Tuple
import numpy as np

from astropy.io import fits
from astropy.io.fits.hdu.image import PrimaryHDU
from astropy.io.fits.hdu.table import BinTableHDU
from astropy.io.fits.hdu.hdulist import HDUList

from spectre_core.spectrograms import Spectrogram
from spectre_core.capture_configs import CaptureModes
from .._register import register_chunk
from .._base import BaseChunk, ChunkFile

@register_chunk(CaptureModes.FIXED_CENTER_FREQUENCY)
class _Chunk(BaseChunk):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs) 
        
        self.add_file(BinChunk(self.chunk_parent_path, self.chunk_name))
        self.add_file(FitsChunk(self.chunk_parent_path, self.chunk_name))
        self.add_file(HdrChunk(self.chunk_parent_path, self.chunk_name))


class BinChunk(ChunkFile):
    def __init__(self, chunk_parent_path: str, chunk_name: str):
        super().__init__(chunk_parent_path, chunk_name, "bin")

    def read(self) -> np.ndarray:
        with open(self.file_path, "rb") as fh:
            return np.fromfile(fh, dtype=np.complex64)


  
class HdrChunk(ChunkFile):
    def __init__(self, chunk_parent_path: str, chunk_name: str):
        super().__init__(chunk_parent_path, chunk_name, "hdr")


    def read(self) -> int:
        hdr_contents = self._extract_contents()
        return self._get_millisecond_correction(hdr_contents)


    def _extract_contents(self) -> np.ndarray:
        with open(self.file_path, "rb") as fh:
            return np.fromfile(fh, dtype=np.float32)


    def _get_millisecond_correction(self, hdr_contents: np.ndarray) -> int:
        if len(hdr_contents) != 1:
            raise ValueError(f"Expected exactly one integer in the header, but received header contents: {hdr_contents}")
        
        millisecond_correction_as_float = float(hdr_contents[0])

        if not millisecond_correction_as_float.is_integer():
            raise TypeError(f"Expected integer value for millisecond correction, but got {millisecond_correction_as_float}")
        
        return int(millisecond_correction_as_float)
        



class FitsChunk(ChunkFile):
    def __init__(self, chunk_parent_path: str, chunk_name: str):
        super().__init__(chunk_parent_path, chunk_name, "fits")


    @property
    def datetimes(self) -> np.ndarray:
        with fits.open(self.file_path, mode='readonly') as hdulist:
            bintable_data = hdulist[1].data
            times = bintable_data['TIME'][0]
            return [self.chunk_start_datetime + timedelta(seconds=t) for t in times]
        

    def read(self) -> Spectrogram:
        with fits.open(self.file_path, mode='readonly') as hdulist:
            primary_hdu = self._get_primary_hdu(hdulist)
            dynamic_spectra = self._get_dynamic_spectra(primary_hdu)
            spectrum_type = self._get_spectrum_type(primary_hdu)
            microsecond_correction = self._get_microsecond_correction(primary_hdu)
            bintable_hdu = self._get_bintable_hdu(hdulist)
            times, frequencies = self._get_time_and_frequency(bintable_hdu)

        return Spectrogram(dynamic_spectra, 
                           times, 
                           frequencies, 
                           self.tag, 
                           chunk_start_time=self.chunk_start_time, 
                           microsecond_correction=microsecond_correction,
                           spectrum_type = spectrum_type)


    def _get_primary_hdu(self, hdulist: HDUList) -> PrimaryHDU:
        return hdulist['PRIMARY']


    def _get_dynamic_spectra(self, primary_hdu: PrimaryHDU) -> np.ndarray:
        return primary_hdu.data


    def _get_spectrum_type(self, primary_hdu: PrimaryHDU) -> str:
        return primary_hdu.header.get('BUNIT', None)


    def _get_microsecond_correction(self, primary_hdu: PrimaryHDU) -> int:
        date_obs = primary_hdu.header.get('DATE-OBS', None)
        time_obs = primary_hdu.header.get('TIME-OBS', None)
        datetime_obs = datetime.strptime(f"{date_obs}T{time_obs}", "%Y-%m-%dT%H:%M:%S.%f")
        return datetime_obs.microsecond


    def _get_bintable_hdu(self, hdulist: HDUList) -> BinTableHDU:
        return hdulist[1]


    def _get_time_and_frequency(self, bintable_hdu: BinTableHDU) -> Tuple[np.ndarray, np.ndarray]:
        data = bintable_hdu.data
        times = data['TIME'][0]
        frequencies_MHz = data['FREQUENCY'][0]
        frequencies = frequencies_MHz * 1e6 # convert to Hz
        return times, frequencies




# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from ._analytical import (
    get_analytical_spectrogram, validate_analytically, TestResults
)
from ._spectrogram import (
    Spectrogram, FrequencyCut, TimeCut, SpectrumTypes, TimeTypes
)
from ._transform import (
    frequency_chop, time_chop, frequency_average, time_average, 
    join_spectrograms
)

__all__ = [
    "get_analytical_spectrogram", "validate_analytically", "TestResults",
    "Spectrogram", "FrequencyCut", "TimeCut", "SpectrumTypes", "frequency_chop",
    "time_chop", "frequency_average","time_average", "join_spectrograms",
    "TimeTypes"
]


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

# decorators run on import
from .library._fixed_center_frequency import _Chunk
from .library._swept_center_frequency import _Chunk
from .library._callisto import _Chunk

from ._base import BaseChunk, ChunkFile
from ._factory import get_chunk_from_tag
from ._chunks import Chunks
from .library._swept_center_frequency import SweepMetadata

__all__ = [
    "BaseChunk",
    "ChunkFile",
    "get_chunk_from_tag",
    "Chunks",
    "SweepMetadata"
]




# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

import os
from typing import Optional
from collections import OrderedDict
import warnings
from datetime import datetime

from spectre_core.spectrograms import Spectrogram, time_chop, join_spectrograms
from spectre_core.config import get_chunks_dir_path, TimeFormats
from spectre_core.exceptions import (
    SpectrogramNotFoundError,
    ChunkNotFoundError
)
from ._base import BaseChunk
from ._factory import get_chunk_from_tag

class Chunks:
    def __init__(self, 
                 tag: str,
                 year: Optional[int] = None, 
                 month: Optional[int] = None, 
                 day: Optional[int] = None):
        self._tag = tag
        self._Chunk = get_chunk_from_tag(tag)
        self._chunk_map: dict[str, BaseChunk] = OrderedDict()
        self._chunk_list: list[BaseChunk] = []
        self._chunk_names: list[str] = []
        self.set_date(year, month, day)


    @property
    def year(self) -> int:
        return self._year


    @property 
    def month(self) -> int:
        return self._month
    

    @property
    def day(self) -> int:
        return self._day


    @property
    def chunks_dir_path(self) -> str:
        return get_chunks_dir_path(self.year, self.month, self.day)
    

    @property 
    def chunk_map(self) -> dict[str, BaseChunk]:
        return self._chunk_map
    

    @property
    def chunk_list(self) -> list[BaseChunk]:
        return  self._chunk_list
    

    @property
    def chunk_names(self) -> list[str]:
        return self._chunk_names


    @property
    def num_chunks(self) -> int:
        return len(self.chunk_list)


    def set_date(self, 
                 year: Optional[int],
                 month: Optional[int],
                 day: Optional[int]) -> None:
        self._year = year
        self._month = month
        self._day = day
        self._update_chunk_map()


    def _update_chunk_map(self) -> None:
        self._chunk_map = OrderedDict() # reset cache
        self._chunk_list = [] # reset cache
        self._chunk_names = [] # reset cache

        chunk_files = [f for (_, _, files) in os.walk(self.chunks_dir_path) for f in files]
        
        if len(chunk_files) == 0:
            warning_message = "No chunks found, setting chunk map with empty dictionary."
            _LOGGER.warning(warning_message)
            warnings.warn(warning_message)
            return
        
        for chunk_file in chunk_files:
            file_name, _ = os.path.splitext(chunk_file)
            chunk_start_time, tag = file_name.split("_", 1)
            if tag == self._tag:
                self._chunk_map[chunk_start_time] = self._Chunk(chunk_start_time, tag)
        
        self._chunk_map = OrderedDict(sorted(self._chunk_map.items()))
        self._chunk_names = list(self._chunk_map.keys())
        self._chunk_list = list(self._chunk_map.values())


    def update(self) -> None:
        """Public alias for setting chunk map"""
        self._update_chunk_map()
    

    def __iter__(self):
        yield from self.chunk_list


    def _get_chunk_by_chunk_start_time(self, 
                                      chunk_start_time: str) -> BaseChunk:
        try:
            return self.chunk_map[chunk_start_time]
        except KeyError:
            raise ChunkNotFoundError(f"Chunk with chunk start time {chunk_start_time} could not be found within {self.chunks_dir_path}")


    def _get_chunk_by_index(self, 
                           chunk_index: int) -> BaseChunk:
        num_chunks = len(self.chunk_map)
        if num_chunks == 0:
            raise ChunkNotFoundError("No chunks are available")
        index = chunk_index % num_chunks  # Use modulo to make the index wrap around. Allows the user to iterate over all the chunks via index cyclically.
        return self.chunk_list[index]


    def __getitem__(self, subscript: str | int):
        if isinstance(subscript, str):
            return self._get_chunk_by_chunk_start_time(subscript)
        elif isinstance(subscript, int):
            return self._get_chunk_by_index(subscript)


    def get_index_by_chunk(self, 
                           chunk_to_match: BaseChunk) -> int:
        for i, chunk in enumerate(self):
            if chunk.chunk_start_time == chunk_to_match.chunk_start_time:
                return i
        raise ChunkNotFoundError(f"No matching chunk found for chunk {chunk_to_match.chunk_name}")
    

    def count_chunk_files(self, 
                          extension: str) -> int:
        return sum(1 for chunk_file in self if chunk_file.has_file(extension))


    def get_spectrogram_from_range(self, 
                                   start_time: str, 
                                   end_time: str) -> Spectrogram:
        # Convert input strings to datetime objects
        start_datetime = datetime.strptime(start_time, TimeFormats.DATETIME)
        end_datetime = datetime.strptime(end_time, TimeFormats.DATETIME)

        if start_datetime.day != end_datetime.day:
            warning_message = "Joining spectrograms across multiple days"
            _LOGGER.warning(warning_message)
            warnings.warn(warning_message, RuntimeWarning)

        spectrograms = []
        num_fits_chunks = self.count_chunk_files("fits")

        for i, chunk in enumerate(self):
            # skip chunks without fits files
            if not chunk.has_file("fits"):
                continue
            
            # rather than reading all files to evaluate the actual upper bound to their time range (slow)
            # place an upper bound by using the chunk start datetime for the next chunk
            # this assumes that the chunks are non-overlapping (reasonable assumption)
            lower_bound = chunk.chunk_start_datetime
            if i < num_fits_chunks:
                next_chunk = self[i + 1]
                upper_bound = next_chunk.chunk_start_datetime
            # if there is no "next chunk" then we do have to read the file
            else:
                fits_chunk = chunk.get_file("fits")
                upper_bound = fits_chunk.datetimes[-1]

            # if the chunk overlaps with the input time range, then read the fits file
            if start_datetime <= upper_bound and lower_bound <= end_datetime:
                spectrogram = chunk.read_file("fits")
                spectrogram = time_chop(spectrogram, start_time, end_time)
                # if we have a non-empty spectrogram, append it to the list of spectrograms
                if spectrogram:
                    spectrograms.append(spectrogram)

        if spectrograms:
            return join_spectrograms(spectrograms)
        else:
            raise SpectrogramNotFoundError("No spectrogram data found for the given time range")



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from ._register import event_handler_map
from spectre_core.post_processing._base import BaseEventHandler
from spectre_core.capture_configs import CaptureConfig, PNames
from spectre_core.exceptions import EventHandlerNotFoundError


def _get_event_handler(event_handler_key: str) -> BaseEventHandler:
    EventHandler = event_handler_map.get(event_handler_key)
    if EventHandler is None:
        valid_event_handler_keys = list(event_handler_map.keys())
        raise EventHandlerNotFoundError((f"No event handler found for the event handler key '{event_handler_key}'. "
                                         f"Please specify one of the following event handler keys: {valid_event_handler_keys}"))
    return EventHandler


def get_event_handler_from_tag(tag: str) -> BaseEventHandler:
    capture_config = CaptureConfig(tag)
    event_handler_key = capture_config.get_parameter_value(PNames.EVENT_HANDLER_KEY)
    return _get_event_handler(event_handler_key)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

from typing import Optional
from abc import ABC, abstractmethod
from scipy.signal import ShortTimeFFT, get_window

from watchdog.events import FileSystemEventHandler, FileCreatedEvent

from spectre_core.capture_configs import CaptureConfig, PNames
from spectre_core.chunks import BaseChunk, get_chunk_from_tag
from spectre_core.spectrograms import Spectrogram, join_spectrograms


def make_sft_instance(capture_config: CaptureConfig
) -> ShortTimeFFT:
    sample_rate   = capture_config.get_parameter_value(PNames.SAMPLE_RATE)
    window_hop    = capture_config.get_parameter_value(PNames.WINDOW_HOP)
    window_type   = capture_config.get_parameter_value(PNames.WINDOW_TYPE)
    window_size   = capture_config.get_parameter_value(PNames.WINDOW_SIZE)
    window = get_window(window_type, 
                        window_size)
    return ShortTimeFFT(window, 
                        window_hop,
                        sample_rate, 
                        fft_mode = "centered")


class BaseEventHandler(ABC, FileSystemEventHandler):
    def __init__(self, 
                 tag: str):
        self._tag = tag

        # the tag tells us 'what type' of data is stored in the files for each chunk
        self._Chunk: BaseChunk = get_chunk_from_tag(tag)
        # load the capture config corresponding to the tag
        self._capture_config   = CaptureConfig(tag)

        # post processing is triggered by files with this extension
        self._watch_extension = self._capture_config.get_parameter_value(PNames.WATCH_EXTENSION)

        # store the next file to be processed (specifically, the absolute file path of the file)
        self._queued_file: Optional[str] = None

        # store batched spectrograms as they are created into a cache
        # which is flushed periodically according to a user defined 
        # time range
        self._cached_spectrogram: Optional[Spectrogram] = None


    @abstractmethod
    def process(self, 
                absolute_file_path: str) -> None:
        """Process the file stored at the input absolute file path.
        
        To be implemented by derived classes.
        """

    def on_created(self, 
                   event: FileCreatedEvent):
        """Process a newly created batch file, only once the next batch is created.
        
        Since we assume that the batches are non-overlapping in time, this guarantees
        we avoid post processing a file while it is being written to. Files are processed
        sequentially, in the order they are created.
        """

        # the 'src_path' attribute holds the absolute path of the newly created file
        absolute_file_path = event.src_path
        
        # only 'notice' a file if it ends with the appropriate extension
        # as defined in the capture config
        if absolute_file_path.endswith(self._watch_extension):
            _LOGGER.info(f"Noticed {absolute_file_path}")
            
            # If there exists a queued file, try and process it
            if self._queued_file is not None:
                try:
                    self.process(self._queued_file)
                except Exception:
                    _LOGGER.error(f"An error has occured while processing {self._queued_file}",
                                  exc_info=True)
                     # flush any internally stored spectrogram on error to avoid lost data
                    self._flush_cache()
                    # re-raise the exception to the main thread
                    raise
            
            # Queue the current file for processing next
            _LOGGER.info(f"Queueing {absolute_file_path} for post processing")
            self._queued_file = absolute_file_path
    

    def _cache_spectrogram(self, 
                           spectrogram: Spectrogram) -> None:
        _LOGGER.info("Joining spectrogram")

        if self._cached_spectrogram is None:
            self._cached_spectrogram = spectrogram
        else:
            self._cached_spectrogram = join_spectrograms([self._cached_spectrogram, spectrogram])

        # if the time range is not specified
        time_range = self._capture_config.get_parameter_value(PNames.TIME_RANGE) or 0.0
  
        if self._cached_spectrogram.time_range >= time_range:
            self._flush_cache()
    

    def _flush_cache(self) -> None:
        if self._cached_spectrogram:
            _LOGGER.info(f"Flushing spectrogram to file with chunk start time "
                         f"'{self._cached_spectrogram.chunk_start_time}'")
            self._cached_spectrogram.save()
            _LOGGER.info("Flush successful, resetting spectrogram cache")
            self._cached_spectrogram = None # reset the cache


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime
from typing import Optional

from spectre_core._file_io import BaseFileHandler
from spectre_core.config import get_chunks_dir_path, TimeFormats
from spectre_core.exceptions import ChunkFileNotFoundError

class ChunkFile(BaseFileHandler):
    def __init__(self, 
                 chunk_parent_path: str, 
                 chunk_name: str, 
                 extension: str, 
                 **kwargs):
        self._chunk_start_time, self._tag = chunk_name.split("_")
        self._chunk_start_datetime: Optional[datetime] = None
        super().__init__(chunk_parent_path, 
                         chunk_name, 
                         extension = extension, 
                         **kwargs)


    @property
    def chunk_start_time(self) -> str:
        return self._chunk_start_time
    

    @property
    def tag(self) -> str:
        return self._tag
    

    @property
    def chunk_start_datetime(self) -> datetime:
        if self._chunk_start_datetime is None:
            self._chunk_start_datetime = datetime.strptime(self.chunk_start_time, TimeFormats.DATETIME)
        return self._chunk_start_datetime
    


class BaseChunk:
    def __init__(self, 
                 chunk_start_time: str, 
                 tag: str):
        self._chunk_start_time: str = chunk_start_time
        self._tag: str = tag
        self._chunk_files: dict[str, ChunkFile] = {}
        self._chunk_start_datetime: Optional[datetime] = None
        self.chunk_parent_path: str = get_chunks_dir_path(year = self.chunk_start_datetime.year,
                                                          month = self.chunk_start_datetime.month,
                                                          day = self.chunk_start_datetime.day)
        self._chunk_name: str = f"{self.chunk_start_time}_{self.tag}"


    @property
    def chunk_start_time(self) -> str:
        return self._chunk_start_time


    @property
    def tag(self) -> str:
        return self._tag
      

    @property
    def chunk_start_datetime(self) -> datetime:
        if self._chunk_start_datetime is None:
            self._chunk_start_datetime = datetime.strptime(self.chunk_start_time, TimeFormats.DATETIME)
        return self._chunk_start_datetime


    @property
    def chunk_name(self) -> str:
        return f"{self._chunk_start_time}_{self._tag}"
    

    @property
    def extensions(self) -> list[str]:
        return list(self._chunk_files.keys())
    
    
    def add_file(self, chunk_file: ChunkFile) -> None:
        self._chunk_files[chunk_file.extension] = chunk_file
    

    def get_file(self, extension: str) -> ChunkFile:
        try:
            return self._chunk_files[extension]
        except KeyError:
            raise ChunkFileNotFoundError(f"No chunk file found with extension '{extension}'")


    def read_file(self, extension: str):
        chunk_file = self.get_file(extension)
        return chunk_file.read()


    def delete_file(self, extension: str, **kwargs):
        chunk_file = self.get_file(extension)
        try:
            chunk_file.delete(**kwargs)
        except FileNotFoundError as e:
            raise ChunkFileNotFoundError(str(e))


    def has_file(self, extension: str) -> bool:
        try:
            chunk_file = self.get_file(extension)
            return chunk_file.exists
        except ChunkFileNotFoundError:
            return False





# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

import os
from typing import Optional
from typing import Tuple
from datetime import timedelta, datetime
import numpy as np

from scipy.signal import ShortTimeFFT

from spectre_core.spectrograms import Spectrogram, time_average, frequency_average
from spectre_core.config import TimeFormats
from spectre_core.capture_configs import CaptureConfig, PNames, CaptureModes
from spectre_core.chunks import BaseChunk
from spectre_core.chunks import SweepMetadata
from spectre_core.exceptions import InvalidSweepMetadataError
from .._base import BaseEventHandler, make_sft_instance
from .._register import register_event_handler


def _stitch_steps(stepped_dynamic_spectra: np.ndarray,
                  num_full_sweeps: int) -> np.ndarray:
    """For each full sweep, create a swept spectrum by stitching together the spectrum at each of the steps."""
    return stepped_dynamic_spectra.reshape((num_full_sweeps, -1)).T

    
def _average_over_steps(stepped_dynamic_spectra: np.ndarray) -> None:
    """Average the spectrums in each step totally in time."""
    return np.nanmean(stepped_dynamic_spectra[..., 1:], axis=-1)


def _fill_times(times: np.ndarray,
                num_samples: np.ndarray,
                sample_rate: int,
                num_full_sweeps: int,
                num_steps_per_sweep: int) -> None:
        """Assign physical times to each swept spectrum. We use (by convention) the time of the first sample in eahc sweep"""

        sampling_interval = 1 / sample_rate
        cumulative_samples = 0
        for sweep_index in range(num_full_sweeps):
            # assign a physical time to the spectrum for this sweep
            times[sweep_index] = cumulative_samples * sampling_interval

            # find the total number of samples across the sweep
            start_step = sweep_index * num_steps_per_sweep
            end_step = (sweep_index + 1) * num_steps_per_sweep

            # update cumulative samples
            cumulative_samples += np.sum(num_samples[start_step:end_step])


def _fill_frequencies(frequencies: np.ndarray,
                      center_frequencies: np.ndarray,
                      baseband_frequencies: np.ndarray,
                      window_size: int) -> None:
    """Assign physical frequencies to each of the swept spectral components."""
    for i, center_frequency in enumerate(np.unique(center_frequencies)):
        lower_bound = i * window_size
        upper_bound = (i + 1) * window_size
        frequencies[lower_bound:upper_bound] = (baseband_frequencies + center_frequency)


def _fill_stepped_dynamic_spectra(stepped_dynamic_spectra: np.ndarray,
                                  sft: ShortTimeFFT,
                                  iq_data: np.ndarray,
                                  num_samples: np.ndarray,
                                  num_full_sweeps: int,
                                  num_steps_per_sweep: int) -> None:
    """Compute the dynamic spectra for the input IQ samples for each step.
    
    All IQ samples per step were collected at the same center frequency.
    """
    # global_step_index will hold the step index over all sweeps (doesn't reset each sweep)
    # start_sample_index will hold the index of the first sample in the step
    global_step_index, start_sample_index = 0, 0
    for sweep_index in range(num_full_sweeps):
        for step_index in range(num_steps_per_sweep):
            # extract how many samples are in the current step from the metadata
            end_sample_index = start_sample_index + num_samples[global_step_index]
            # compute the number of slices in the current step based on the window we defined on the capture config
            num_slices = sft.upper_border_begin(num_samples[global_step_index])[1]
            # perform a short time fast fourier transform on the step
            complex_spectra = sft.stft(iq_data[start_sample_index:end_sample_index], 
                                            p0=0, 
                                            p1=num_slices)
            # and pack the absolute values into the stepped spectrogram where the step slot is padded to the maximum size for ease of processing later)
            stepped_dynamic_spectra[sweep_index, step_index, :, :num_slices] = np.abs(complex_spectra)
            # reassign the start_sample_index for the next step
            start_sample_index = end_sample_index
            # and increment the global step index
            global_step_index += 1
        

def _compute_num_max_slices_in_step(sft: ShortTimeFFT,
                                    num_samples: np.ndarray) -> int:
    """Compute the maximum number of slices over all steps, in all sweeps over the chunk."""
    return sft.upper_border_begin(np.max(num_samples))[1]


def _compute_num_full_sweeps(center_frequencies: np.ndarray) -> int:
    """Compute the total number of full sweeps over the chunk.

    Since the number of each samples in each step is variable, we only know a sweep is complete
    when there is a sweep after it. So we can define the total number of *full* sweeps as the number of 
    (freq_max, freq_min) pairs in center_frequencies. It is only at an instance of (freq_max, freq_min) pair 
    in center frequencies that the frequency decreases, so, we can compute the number of full sweeps by 
    counting the numbers of negative values in np.diff(center_frequencies)
    """
    return len(np.where(np.diff(center_frequencies) < 0)[0])


def _compute_num_steps_per_sweep(center_frequencies: np.ndarray) -> int:
    """Compute the (ensured constant) number of steps in each sweep."""
    # find the (step) indices corresponding to the minimum frequencies
    min_freq_indices = np.where(center_frequencies == np.min(center_frequencies))[0]
    # then, we evaluate the number of steps that has occured between them via np.diff over the indices
    unique_num_steps_per_sweep = np.unique(np.diff(min_freq_indices))
    # we expect that the difference is always the same, so that the result of np.unique has a single element
    if len(unique_num_steps_per_sweep) != 1:
        raise InvalidSweepMetadataError(("Irregular step count per sweep, "
                                         "expected a consistent number of steps per sweep"))
    return int(unique_num_steps_per_sweep[0])


def _validate_center_frequencies_ordering(center_frequencies: np.ndarray,
                                          freq_step: float) -> None:
    """Check that the center frequencies are well-ordered in the detached header"""
    min_frequency = np.min(center_frequencies)
    # Extract the expected difference between each step within a sweep. 
    for i, diff in enumerate(np.diff(center_frequencies)):
        # steps should either increase by freq_step or drop to the minimum
        if (diff != freq_step) and (center_frequencies[i + 1] != min_frequency):
            raise InvalidSweepMetadataError(f"Unordered center frequencies detected")


def _do_stfft(iq_data: np.ndarray,
              sweep_metadata: SweepMetadata,
              capture_config: CaptureConfig):
    """Perform a Short Time FFT on the input swept IQ samples."""

    sft = make_sft_instance(capture_config)

    frequency_step = capture_config.get_parameter_value(PNames.FREQUENCY_STEP)
    _validate_center_frequencies_ordering(sweep_metadata.center_frequencies,
                                          frequency_step)

    window_size = capture_config.get_parameter_value(PNames.WINDOW_SIZE)

    num_steps_per_sweep = _compute_num_steps_per_sweep(sweep_metadata.center_frequencies)
    num_full_sweeps = _compute_num_full_sweeps(sweep_metadata.center_frequencies)
    num_max_slices_in_step = _compute_num_max_slices_in_step(sft,
                                                             sweep_metadata.num_samples)
    
    stepped_dynamic_spectra_shape = (num_full_sweeps, 
                                     num_steps_per_sweep, 
                                     window_size, 
                                     num_max_slices_in_step)
    stepped_dynamic_spectra = np.full(stepped_dynamic_spectra_shape, np.nan)

    frequencies = np.empty(num_steps_per_sweep * window_size)
    times = np.empty(num_full_sweeps)

    _fill_stepped_dynamic_spectra(stepped_dynamic_spectra,
                                  sft,
                                  iq_data,
                                  sweep_metadata.num_samples,
                                  num_full_sweeps,
                                  num_steps_per_sweep)
    
    _fill_frequencies(frequencies,
                      sweep_metadata.center_frequencies,
                      sft.f,
                      window_size)
    
    sample_rate = capture_config.get_parameter_value(PNames.SAMPLE_RATE)
    _fill_times(times,
                sweep_metadata.num_samples,
                sample_rate,
                num_full_sweeps,
                num_steps_per_sweep)

    averaged_spectra = _average_over_steps(stepped_dynamic_spectra)
    dynamic_spectra = _stitch_steps(averaged_spectra,
                                    num_full_sweeps)

    return times, frequencies, dynamic_spectra


def _correct_timing(chunk_start_datetime: datetime,
                    millisecond_correction: int,
                    num_samples_prepended: int,
                    sample_rate: int):
    """Correct the start time for this chunk based on the number of samples we prepended reconstructing the initial sweep."""
    sample_interval = (1 / sample_rate)
    elapsed_time = num_samples_prepended * sample_interval
    corrected_datetime = chunk_start_datetime + timedelta(milliseconds = millisecond_correction) - timedelta(seconds = float(elapsed_time))
    return corrected_datetime.strftime(TimeFormats.DATETIME), corrected_datetime.microsecond * 1e-3


def _prepend_num_samples(carryover_num_samples: np.ndarray,
                         num_samples: np.ndarray,
                         final_step_spans_two_chunks: bool) -> np.ndarray:
    """Prepend the number of samples from the final sweep of the previous chunk."""
    if final_step_spans_two_chunks:
        # ensure the number of samples from the final step in the previous chunk are accounted for
        num_samples[0] += carryover_num_samples[-1]
        # and truncate as required
        carryover_num_samples = carryover_num_samples[:-1]
    return np.concatenate((carryover_num_samples, num_samples))


def _prepend_center_frequencies(carryover_center_frequencies: np.ndarray,
                                center_frequencies: np.ndarray,
                                final_step_spans_two_chunks: bool)-> np.ndarray:
    """Prepend the center frequencies from the final sweep of the previous chunk."""
    # in the case that the sweep has bled across chunks,
    # do not permit identical neighbours in the center frequency array
    if final_step_spans_two_chunks:
        # truncate the final frequency to prepend (as it already exists in the array we are appending to in this case)
        carryover_center_frequencies = carryover_center_frequencies[:-1]
    return np.concatenate((carryover_center_frequencies, center_frequencies))


def _prepend_iq_data(carryover_iq_data: np.ndarray,
                     iq_data: np.ndarray) -> np.ndarray:
    """Prepend the IQ samples from the final sweep of the previous chunk."""
    return np.concatenate((carryover_iq_data, iq_data))


def _get_final_sweep(previous_chunk: BaseChunk
) -> Tuple[np.ndarray, SweepMetadata]:
    """Get data from the final sweep of the previous chunk."""
    # unpack the data from the previous chunk
    previous_iq_data = previous_chunk.read_file("bin")
    _, previous_sweep_metadata = previous_chunk.read_file("hdr")
    # find the step index from the last sweep
    # [0] since the return of np.where is a 1 element Tuple, 
    # containing a list of step indices corresponding to the smallest center frequencies
    # [-1] since we want the final step index, where the center frequency is minimised
    final_sweep_start_step_index = np.where(previous_sweep_metadata.center_frequencies == np.min(previous_sweep_metadata.center_frequencies))[0][-1]
    # isolate the data from the final sweep
    final_center_frequencies = previous_sweep_metadata.center_frequencies[final_sweep_start_step_index:]
    final_num_samples = previous_sweep_metadata.num_samples[final_sweep_start_step_index:]
    final_sweep_iq_data = previous_iq_data[-np.sum(final_num_samples):]

    # sanity check on the number of samples in the final sweep
    if len(final_sweep_iq_data) != np.sum(final_num_samples):
        raise ValueError((f"Unexpected error! Mismatch in sample count for the final sweep data."
                            f"Expected {np.sum(final_num_samples)} based on sweep metadata, but found "
                            f" {len(final_sweep_iq_data)} IQ samples in the final sweep"))
    
    return final_sweep_iq_data, SweepMetadata(final_center_frequencies, final_num_samples)


def _reconstruct_initial_sweep(previous_chunk: BaseChunk,
                               iq_data: np.ndarray,
                               sweep_metadata: SweepMetadata) -> Tuple[np.ndarray, SweepMetadata, int]:
    """Reconstruct the initial sweep of the current chunk, using data from the previous chunk."""

    # carryover the final sweep of the previous chunk, and prepend that data to the current chunk data
    carryover_iq_data, carryover_sweep_metadata = _get_final_sweep(previous_chunk)

    # prepend the iq data that was carried over from the previous chunk
    iq_data = _prepend_iq_data(carryover_iq_data,
                               iq_data)
    
    # prepend the sweep metadata from the previous chunk
    final_step_spans_two_chunks = carryover_sweep_metadata.center_frequencies[-1] == sweep_metadata.center_frequencies[0]
    center_frequencies = _prepend_center_frequencies(carryover_sweep_metadata.center_frequencies,
                                                     sweep_metadata.center_frequencies,
                                                     final_step_spans_two_chunks)
    num_samples = _prepend_num_samples(carryover_sweep_metadata.num_samples,
                                       sweep_metadata.num_samples,
                                       final_step_spans_two_chunks)
    
    # keep track of how many samples we prepended (required to adjust timing later)
    num_samples_prepended = np.sum(carryover_sweep_metadata.num_samples)
    return iq_data, SweepMetadata(center_frequencies, num_samples), num_samples_prepended


def _build_spectrogram(chunk: BaseChunk,
                       capture_config: CaptureConfig,
                       previous_chunk: Optional[BaseChunk] = None) -> Spectrogram:
    """Create a spectrogram by performing a Short Time FFT on the (swept) IQ samples for this chunk."""
    iq_data = chunk.read_file("bin")
    millisecond_correction, sweep_metadata = chunk.read_file("hdr")

    # if a previous chunk has been specified, this indicates that the initial sweep spans
    # between two adjacent batched files. 
    if previous_chunk:
        # If this is the case, first reconstruct the initial sweep of the current chunk
        # by prepending the final sweep of the previous chunk
        iq_data, sweep_metadata, num_samples_prepended = _reconstruct_initial_sweep(previous_chunk,
                                                                                    iq_data,
                                                                                    sweep_metadata)
        # since we have prepended extra samples, we need to correct the chunk start time
        # appropriately
        chunk_start_time, millisecond_correction = _correct_timing(chunk.chunk_start_datetime,
                                                                   millisecond_correction,
                                                                   num_samples_prepended,
                                                                   capture_config.get_parameter_value(PNames.SAMPLE_RATE))
    # otherwise, no action is required
    else:
        chunk_start_time = chunk.chunk_start_time

    microsecond_correction = millisecond_correction * 1e3

    times, frequencies, dynamic_spectra = _do_stfft(iq_data, 
                                                    sweep_metadata,
                                                    capture_config)
    
    return Spectrogram(dynamic_spectra,
                       times,
                       frequencies,
                       chunk.tag,
                       chunk_start_time,
                       microsecond_correction,
                       spectrum_type = "amplitude")


@register_event_handler(CaptureModes.SWEPT_CENTER_FREQUENCY)
class _EventHandler(BaseEventHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # the previous chunk is stored in order to fetch the
        # data from the "final sweep" which was ignored during
        # processing.
        self._previous_chunk: BaseChunk = None 
        

    def process(self, 
                absolute_file_path: str):
        _LOGGER.info(f"Processing: {absolute_file_path}")
        file_name = os.path.basename(absolute_file_path)
        # discard the extension
        base_file_name, _ = os.path.splitext(file_name)
        chunk_start_time, tag = base_file_name.split('_')
        chunk = self._Chunk(chunk_start_time, tag)

        # ensure that the file which has been created has the expected tag
        if tag != self._tag:
            raise RuntimeError(f"Received an unexpected tag! Expected '{self._tag}', "
                               f"but a file has been created with tag '{tag}'")

        _LOGGER.info("Creating spectrogram")
        spectrogram = _build_spectrogram(chunk,
                                        self._capture_config,
                                        previous_chunk = self._previous_chunk)

        spectrogram = time_average(spectrogram,
                                   resolution = self._capture_config.get_parameter_value(PNames.TIME_RESOLUTION))

        spectrogram = frequency_average(spectrogram,
                                        resolution = self._capture_config.get_parameter_value(PNames.FREQUENCY_RESOLUTION))

        self._cache_spectrogram(spectrogram)

        # if the previous chunk has not yet been set, it means we are processing the first chunk
        # so we don't need to handle the previous chunk
        if self._previous_chunk is None:
            # instead, only set it for the next time this method is called
            self._previous_chunk = chunk
            
        # otherwise the previous chunk is defined (and by this point has already been processed)
        else:
            bin_chunk = self._previous_chunk.get_file('bin')
            _LOGGER.info(f"Deleting {bin_chunk.file_path}")
            bin_chunk.delete()

            hdr_chunk = self._previous_chunk.get_file('hdr')
            _LOGGER.info(f"Deleting {hdr_chunk.file_path}")
            hdr_chunk.delete()

            # and reassign the current chunk to be used as the previous chunk at the next call of this method
            self._previous_chunk = chunk



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod
from typing import Optional
import numpy as np
from dataclasses import dataclass

from matplotlib import cm
import matplotlib.dates as mdates
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from spectre_core.spectrograms import Spectrogram, TimeTypes
from ._format import PanelFormat


@dataclass(frozen=True)
class XAxisTypes:
    TIME     : str = "time"
    FREQUENCY: str = "frequency"


class BasePanel(ABC):
    def __init__(self, 
                 name: str,
                 spectrogram: Spectrogram):
        self._name = name
        self._spectrogram = spectrogram

        self._x_axis_type: Optional[str] = None
        self._set_x_axis_type()

        # defined while stacking
        self._panel_format: Optional[PanelFormat] = None
        self._time_type: Optional[str] = None
        self._ax: Optional[Axes] = None
        self._fig: Optional[Figure] = None
         # defined if specified by the user
        self._identifier: Optional[str] = None


    @abstractmethod
    def draw(self):
        pass


    @abstractmethod
    def annotate_x_axis(self):
        pass


    @abstractmethod
    def annotate_y_axis(self):
        pass


    @abstractmethod
    def _set_x_axis_type(self):
        """Required to allow for axes sharing in the stack"""
        pass


    @property
    def spectrogram(self) -> Spectrogram:
        return self._spectrogram
    

    @property
    def tag(self) -> str:
        return self._spectrogram.tag
    

    @property
    def time_type(self) -> str:
        return self._time_type
    

    @time_type.setter
    def time_type(self, value: str) -> None:
        self._validate_time_type(value)
        self._time_type = value
    

    @property
    def name(self) -> str:
        if self._name is None:
            raise AttributeError(f"Name for this panel has not yet been set")
        return self._name
    

    @name.setter
    def name(self, value: str) -> None:
        self._name = value
    
    
    @property
    def panel_format(self) -> PanelFormat:
        if self._panel_format is None:
            raise AttributeError(f"Panel format has not yet been specified for this panel")
        return self._panel_format


    @panel_format.setter
    def panel_format(self, value: PanelFormat):
        self._panel_format = value


    @property
    def ax(self) -> Axes:
        if self._ax is None:
            raise AttributeError(f"Axes have not yet been set for this panel")
        return self._ax
    

    @ax.setter
    def ax(self, value: Axes) -> None:
        self._ax = value


    @property
    def fig(self) -> Figure:
        if self._fig is None:
            raise AttributeError(f"Figure has not yet been set for this panel")
        return self._fig
    

    @fig.setter
    def fig(self, value: Figure) -> None:
        self._fig = value
    


    @property
    def x_axis_type(self) -> str:
        if self._x_axis_type is None:
            raise AttributeError(f"x-axis type has not been defined for this panel")
        return self._x_axis_type
    
    @property
    def identifier(self) -> Optional[str]:
        return self._identifier
    
    
    @identifier.setter
    def identifier(self, value: str) -> None:
        self._identifier = value


    def _validate_time_type(self, 
                            time_type: str):
        valid_time_types = [TimeTypes.SECONDS, TimeTypes.DATETIMES]
        if time_type not in valid_time_types:
            raise ValueError(f"Invalid time type. "
                             f"Expected one of {valid_time_types} "
                             f"but received {time_type}")
    

    def bind_to_colors(self, 
                       values: np.ndarray, 
                       cmap: str = "winter"): 
        cmap = cm.get_cmap(cmap)
        rgbas = cmap(np.linspace(0.1, 0.9, len(values)))
        return zip(values, rgbas) # assign each RGBA array to each value
    

    def hide_x_axis_labels(self) -> None:
        self.ax.tick_params(axis='x', labelbottom=False)


    def hide_y_axis_labels(self) -> None:
        self.ax.tick_params(axis='y', labelbottom=False)
    

class BaseTimeSeriesPanel(BasePanel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    

    @property
    def times(self):
        return self.spectrogram.times if self.time_type == TimeTypes.SECONDS else self.spectrogram.datetimes
    

    def annotate_x_axis(self):
        if self.time_type == TimeTypes.SECONDS:
            self.ax.set_xlabel('Time [s]')
        else:
            self.ax.set_xlabel('Time [UTC]')
            self.ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M:%S'))
        
        
    def _set_x_axis_type(self):
        self._x_axis_type = XAxisTypes.TIME



class BaseSpectrumPanel(BasePanel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    @property
    def frequencies(self):
        return self._spectrogram.frequencies


    def annotate_x_axis(self):
        self.ax.set_xlabel('Frequency [MHz]')


    def _set_x_axis_type(self):
        self._x_axis_type = XAxisTypes.FREQUENCY


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

from dataclasses import dataclass
from math import floor
from scipy.signal import get_window
from numbers import Number
from warnings import warn
from typing import Optional

from ._parameters import Parameters
from ._ptemplates import PNames


def _validate_window(
        parameters: Parameters
) -> None:
    window_size = parameters.get_parameter_value(PNames.WINDOW_SIZE)
    window_type = parameters.get_parameter_value(PNames.WINDOW_TYPE)
    sample_rate = parameters.get_parameter_value(PNames.SAMPLE_RATE)
    batch_size  = parameters.get_parameter_value(PNames.BATCH_SIZE)
    
    window_interval = window_size*(1 / sample_rate)
    if window_interval > batch_size:
        raise ValueError((f"The windowing interval must be strictly less than the chunk size. "
                          f"Computed the windowing interval to be {window_interval} [s], "
                          f"but the chunk size is {batch_size} [s]"))
    
    try:
        _ = get_window(window_type, window_size)
    except Exception as e:
        raise Exception((f"An error has occurred while validating the window. "
                         f"Got {str(e)}"))
    

def _validate_nyquist_criterion(
        parameters: Parameters
) -> None:
    sample_rate = parameters.get_parameter_value(PNames.SAMPLE_RATE)
    bandwidth   = parameters.get_parameter_value(PNames.BANDWIDTH)

    if sample_rate < bandwidth:
        raise ValueError((f"Nyquist criterion has not been satisfied. "
                          f"Sample rate must be greater than or equal to the bandwidth. "
                          f"Got sample rate {sample_rate} [Hz], and bandwidth {bandwidth} [Hz]"))
    

def _compute_num_steps_per_sweep(min_freq: float, 
                                 max_freq: float,
                                 samp_rate: int,
                                 freq_step: float) -> int:
     return floor((max_freq - min_freq + samp_rate/2) / freq_step)


def _validate_num_steps_per_sweep(
        parameters: Parameters
) -> None:
    min_freq    = parameters.get_parameter_value(PNames.MIN_FREQUENCY)
    max_freq    = parameters.get_parameter_value(PNames.MAX_FREQUENCY)
    sample_rate = parameters.get_parameter_value(PNames.SAMPLE_RATE)
    freq_step   = parameters.get_parameter_value(PNames.FREQUENCY_STEP)

    num_steps_per_sweep = _compute_num_steps_per_sweep(min_freq, 
                                                       max_freq, 
                                                       sample_rate, 
                                                       freq_step)
    if num_steps_per_sweep <= 1:
        raise ValueError((f"We need strictly greater than one step per sweep. "
                          f"Computed {num_steps_per_sweep} step per sweep"))
    

def _validate_sweep_interval(
        parameters: Parameters
) -> None: 
    min_freq         = parameters.get_parameter_value(PNames.MIN_FREQUENCY)
    max_freq         = parameters.get_parameter_value(PNames.MAX_FREQUENCY)
    sample_rate      = parameters.get_parameter_value(PNames.SAMPLE_RATE)
    freq_step        = parameters.get_parameter_value(PNames.FREQUENCY_STEP)
    samples_per_step = parameters.get_parameter_value(PNames.SAMPLES_PER_STEP)
    batch_size       = parameters.get_parameter_value(PNames.BATCH_SIZE)

    num_steps_per_sweep = _compute_num_steps_per_sweep(min_freq, 
                                                       max_freq, 
                                                       sample_rate, 
                                                       freq_step)
    num_samples_per_sweep = num_steps_per_sweep * samples_per_step
    sweep_interval = num_samples_per_sweep * 1/sample_rate
    if sweep_interval > batch_size:
        raise ValueError((f"Sweep interval must be less than the chunk size. "
                          f"The computed sweep interval is {sweep_interval} [s], "
                          f"but the given chunk size is {batch_size} [s]"))
    

def _validate_num_samples_per_step(
        parameters: Parameters
) -> None:

    window_size      = parameters.get_parameter_value(PNames.WINDOW_SIZE)
    samples_per_step = parameters.get_parameter_value(PNames.SAMPLES_PER_STEP)

    if window_size >= samples_per_step:
        raise ValueError((f"Window size must be strictly less than the number of samples per step. "
                          f"Got window size {window_size} [samples], which is more than or equal "
                          f"to the number of samples per step {samples_per_step}"))
    

def _validate_non_overlapping_steps(
        parameters: Parameters
) -> None:
    
    freq_step = parameters.get_parameter_value(PNames.FREQUENCY_STEP)
    sample_rate = parameters.get_parameter_value(PNames.SAMPLE_RATE)

    if freq_step < sample_rate:
        raise NotImplementedError(f"SPECTRE does not yet support spectral steps overlapping in frequency. "
                                  f"Got frequency step {freq_step * 1e-6} [MHz] which is less than the sample "
                                  f"rate {sample_rate * 1e-6} [MHz]")
    

def _validate_step_interval(
        parameters: Parameters,
        api_latency: Number
) -> None:
    
    samples_per_step = parameters.get_parameter_value(PNames.SAMPLES_PER_STEP)
    sample_rate      = parameters.get_parameter_value(PNames.SAMPLE_RATE)

    step_interval = samples_per_step * 1/ sample_rate # [s]
    if step_interval < api_latency:
        raise ValueError(f"The computed step interval of {step_interval} [s] is of the order of empirically "
                         f"derived api latency {api_latency} [s]; you may experience undefined behaviour!")


def _validate_fixed_center_frequency_parameters(
    parameters: Parameters
) -> None:
    _validate_nyquist_criterion(parameters)
    _validate_window(parameters)


def _validate_swept_center_frequency_parameters(
    parameters: Parameters,
    api_retuning_latency: Optional[Number] = None,
) -> None:
    _validate_nyquist_criterion(parameters)
    _validate_window(parameters)
    _validate_non_overlapping_steps(parameters)
    _validate_num_steps_per_sweep(parameters)
    _validate_num_samples_per_step(parameters)
    _validate_sweep_interval(parameters)
    
    if api_retuning_latency is not None:
        _validate_step_interval(parameters, api_retuning_latency)
    

@dataclass(frozen=True)
class PValidators:
    window                 = _validate_window
    nyquist_criterion      = _validate_nyquist_criterion
    step_interval          = _validate_step_interval
    non_overlapping_steps  = _validate_non_overlapping_steps
    num_steps_per_sweep    = _validate_num_steps_per_sweep
    num_samples_per_step   = _validate_num_samples_per_step
    sweep_interval         = _validate_sweep_interval
    step_interval          = _validate_step_interval
    fixed_center_frequency = _validate_fixed_center_frequency_parameters
    swept_center_frequency = _validate_swept_center_frequency_parameters


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

#  Global dictionaries to hold the mappings
chunk_map = {}

# classes decorated with @register_chunk([CHUNK_KEY])
# will be added to chunk_map
def register_chunk(chunk_key: str):
    def decorator(cls):
        chunk_map[chunk_key] = cls
        return cls
    return decorator




# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass

@dataclass(frozen=True)
class SpecNames:
    """A centralised store of specification names"""
    FREQUENCY_LOWER_BOUND   : str = "frequency_lower_bound"
    FREQUENCY_UPPER_BOUND   : str = "frequency_upper_bound"
    SAMPLE_RATE_LOWER_BOUND : str = "sample_rate_lower_bound"
    SAMPLE_RATE_UPPER_BOUND : str = "sample_rate_upper_bound"
    BANDWIDTH_LOWER_BOUND   : str = "bandwidth_lower_bound"
    BANDWIDTH_UPPER_BOUND   : str = "bandwidth_upper_bound"
    BANDWIDTH_OPTIONS       : str = "bandwidth_options"
    DEFINED_BANDWIDTHS      : str = "defined_bandwidths"
    IF_GAIN_UPPER_BOUND     : str = "if_gain_upper_bound"
    RF_GAIN_UPPER_BOUND     : str = "rf_gain_upper_bound"
    API_RETUNING_LATENCY    : str = "api_retuning_latency"


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

# event handler class decorators take effect on import
from .library._fixed_center_frequency import _EventHandler
from .library._swept_center_frequency import _EventHandler

from ._factory import get_event_handler_from_tag
from ._post_processor import PostProcessor

__all__ = [
    "PostProcessor", "get_event_handler_from_tag"
]


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

# Global dictionaries to hold the mappings
receivers = {}

# classes decorated with @register_receiver("<receiver_name>")
# will be added to the global map of receivers with key "receiver_name"
def register_receiver(receiver_name: str):
    def decorator(cls):
        receivers[receiver_name] = cls
        return cls
    return decorator

# return a list of all receiver names
def list_all_receiver_names() -> list[str]:
    return list(receivers.keys())







# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Package-wide default datetime formats.
"""

from dataclasses import dataclass

@dataclass(frozen=True)
class TimeFormats:
    TIME     = "%H:%M:%S"
    DATE     = "%Y-%m-%d"
    DATETIME = f"{DATE}T{TIME}"


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from abc import ABC, abstractmethod
from typing import TypeVar, Optional, Any
from numbers import Number

T = TypeVar('T')

class PConstraint(ABC):
    """Abstract base class for parameter constraints."""

    @abstractmethod
    def constrain(self, value: T) -> None:
        """Apply a constraint to the input parameter."""
        pass

    def __format__(self, format_spec: str = "") -> str:
        attrs = ", ".join(f"{key}={value!r}" for key, value in vars(self).items())
        return f"{self.__class__.__name__}({attrs})"

class Bound(PConstraint):
    """Constrain a parameter value to a specific range."""

    def __init__(self,
                 lower_bound: Optional[Number] = None,
                 upper_bound: Optional[Number] = None,
                 strict_lower: bool = False,
                 strict_upper: bool = False):
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self._strict_lower = strict_lower
        self._strict_upper = strict_upper

    def constrain(self, value: Number) -> None:
        if self._lower_bound is not None:
            if self._strict_lower and value <= self._lower_bound:
                raise ValueError(f"Value must be strictly greater than {self._lower_bound}. "
                                 f"Got {value}.")
            if not self._strict_lower and value < self._lower_bound:
                raise ValueError(f"Value must be greater than or equal to {self._lower_bound}. "
                                 f"Got {value}.")

        if self._upper_bound is not None:
            if self._strict_upper and value >= self._upper_bound:
                raise ValueError(f"Value must be strictly less than {self._upper_bound}. "
                                 f"Got {value}.")
            if not self._strict_upper and value > self._upper_bound:
                raise ValueError(f"Value must be less than or equal to {self._upper_bound}. "
                                 f"Got {value}.")


class OneOf(PConstraint):
    """Constrain a parameter to one of a set of defined options."""

    def __init__(self, options: list[Any] = None):
        self._options = options


    def constrain(self, value: Any) -> None:
        if value not in self._options:
            raise ValueError(f"Value must be one of {self._options}. Got {value}.")
        

class _PowerOfTwo(PConstraint):
    """Constrain a parameter value to be a power of two."""

    def constrain(self, value: Number) -> None:
        if value <= 0 or (value & (value - 1)) != 0:
            raise ValueError(f"Value must be a power of two. Got {value}.")


@dataclass(frozen=True)
class PConstraints:
    """Ready-made pconstraint instances for frequent use-cases."""
    power_of_two           = _PowerOfTwo()
    enforce_positive       = Bound(lower_bound=0, strict_lower=True)
    enforce_negative       = Bound(upper_bound=0, strict_upper=True)
    enforce_non_negative   = Bound(lower_bound=0, strict_lower=False)
    enforce_non_positive   = Bound(upper_bound=0, strict_upper=False)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import logging
from datetime import datetime

from spectre_core.config import TimeFormats
from ._log_handlers import LogHandler

def configure_root_logger(process_type: str, 
                          level: int = logging.INFO
) -> LogHandler:
    system_datetime = datetime.now()
    datetime_stamp = system_datetime.strftime(TimeFormats.DATETIME)
    pid = os.getpid()
    log_handler = LogHandler(datetime_stamp, pid, process_type)
    log_handler.make_parent_path()

    # configure the root logger
    logger = logging.getLogger()
    logger.setLevel(level)
    # Remove any existing handlers to avoid duplicate logs
    for handler in logger.handlers:
        logger.removeHandler(handler)
    # Set up file handler with specific filename
    file_handler = logging.FileHandler(log_handler.file_path)
    file_handler.setLevel(level)
    formatter = logging.Formatter("[%(asctime)s] [%(levelname)8s] --- %(message)s (%(name)s:%(lineno)s)")
    file_handler.setFormatter(formatter)
    # and add it to the root logger
    logger.addHandler(file_handler)

    return log_handler


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Optional, TypeVar, Any
from copy import deepcopy
from textwrap import dedent
from dataclasses import dataclass

from ._pconstraints import PConstraint, PConstraints
from ._parameters import Parameter

T = TypeVar('T')

class PTemplate:
    """A parameter template. 
    
    Constrain the value and type that a parameter can take.
    """
    def __init__(self,
                 name: str,
                 ptype: T,
                 default: Optional[T] = None,
                 nullable: bool = False,
                 enforce_default: Optional[bool] = False,
                 help: Optional[str] = None,
                 pconstraints: Optional[list[PConstraint]] = None):
        if not callable(ptype):
            raise TypeError("ptype must be callable.")

        self._name = name
        self._ptype = ptype
        self._default = default
        self._nullable = nullable
        self._enforce_default = enforce_default
        self._help = dedent(help).strip().replace("\n", " ") if help else "No help has been provided."
        self._pconstraints: list[PConstraint] = pconstraints or []


    @property
    def name(self) -> str:
        """The name of the parameter."""
        return self._name
    

    @property
    def ptype(self) -> T:
        """The parameter type."""
        return self._ptype
    

    @property
    def default(self) -> Optional[T]:
        """The value of the parameter, if the value is unspecified."""
        return self._default
    

    @default.setter
    def default(self, value: T) -> None:
        """Update the default of a ptemplate"""
        self._default = value


    @property
    def nullable(self) -> bool:
        """Whether the value of the parameter is allowed to be of None type."""
        return self._nullable
    

    @property
    def enforce_default(self) -> bool:
        """Whether the provided default value is enforced."""
        return self._enforce_default
    

    @enforce_default.setter
    def enforce_default(self, value: bool) -> None:
        """Set whether the provided default value is enforced."""
        self._enforce_default = value
    

    @property
    def help(self) -> str:
        """A description of what the parameter is, and the value it stores."""
        return self._help
    
    
    def add_pconstraint(self,
                        pconstraint: PConstraint) -> None:
        self._pconstraints.append(pconstraint)


    def _cast(self, 
              value: Any) -> T:
        """Cast the input value to the ptype of this template"""
        try:
            return self._ptype(value)
        except (TypeError, ValueError) as e:
            raise ValueError(f"Could not cast '{value}' to '{self._ptype.__name__}': {e}")


    def _constrain(self, 
                   value: T) -> T:
        """Constrain the input value according to constraints of the template."""

        if self._enforce_default and value != self._default:
            raise ValueError(f"The default value of '{self._default}' "
                             f"is required for the parameter '{self._name}'.")

        # apply existing pconstraints
        for constraint in self._pconstraints:
            try:
                constraint.constrain(value)
            except ValueError as e:
                raise ValueError(f"PConstraint '{constraint.__class__.__name__}' failed for the parameter '{self._name}': {e}")
            except Exception as e:
                raise RuntimeError(f"An unexpected error occurred while applying the pconstraint '{constraint.__class__.__name__}' to "
                                    f"'{self.name}': {e}")
        return value


    def apply_template(self, 
                       value: Optional[Any]) -> T:
        """Cast the value and constrain it according to this ptemplate."""
        if value is None:
            if self._default is not None:
                value = self._default
            elif not self._nullable:
                raise ValueError(f"The parameter '{self._name}' is not nullable, "
                                 f"but no value or default has been provided. "
                                 f"Either provide a value, or provide a default.")
            else:
                return None
        
        value = self._cast(value)
        value = self._constrain(value)
        return value


    def make_parameter(self, 
                       value: Optional[Any] = None) -> Parameter:
        value = self.apply_template(value)
        return Parameter(self._name, value)


    def to_dict(self) -> dict:
        """Convert the template to a dictionary representation."""
        return {
            "name": self._name,
            "type": str(self._ptype),
            "default": self._default,
            "enforce_default": self._enforce_default,
            "help": self._help,
            "constraints": [f"{constraint}" for constraint in self._pconstraints]
        }
    

@dataclass(frozen=True)
class PNames:
    """A centralised store of default parameter template names"""
    CENTER_FREQUENCY        : str = "center_frequency"
    MIN_FREQUENCY           : str = "min_frequency"
    MAX_FREQUENCY           : str = "max_frequency"
    FREQUENCY_STEP          : str = "frequency_step"
    FREQUENCY               : str = "frequency"
    BANDWIDTH               : str = "bandwidth"
    SAMPLE_RATE             : str = "sample_rate"
    IF_GAIN                 : str = "if_gain"
    RF_GAIN                 : str = "rf_gain"
    AMPLITUDE               : str = "amplitude"
    FREQUENCY               : str = "frequency"
    TIME_RESOLUTION         : str = "time_resolution"
    FREQUENCY_RESOLUTION    : str = "frequency_resolution"
    TIME_RANGE              : str = "time_range"
    BATCH_SIZE              : str = "batch_size"
    WINDOW_TYPE             : str = "window_type"
    WINDOW_HOP              : str = "window_hop"
    WINDOW_SIZE             : str = "window_size"
    EVENT_HANDLER_KEY       : str = "event_handler_key"
    WATCH_EXTENSION         : str = "watch_extension"
    CHUNK_KEY               : str = "chunk_key"
    SAMPLES_PER_STEP        : str = "samples_per_step"
    MIN_SAMPLES_PER_STEP    : str = "min_samples_per_step"
    MAX_SAMPLES_PER_STEP    : str = "max_samples_per_step"
    STEP_INCREMENT          : str = "step_increment"
    ORIGIN                  : str = "origin"
    TELESCOPE               : str = "telescope"
    INSTRUMENT              : str = "instrument"
    OBJECT                  : str = "object"
    OBS_LAT                 : str = "obs_lat"
    OBS_LON                 : str = "obs_lon"
    OBS_ALT                 : str = "obs_alt"

#
# All stored base ptemplates
#
_base_ptemplates = {
    PNames.CENTER_FREQUENCY:       PTemplate(PNames.CENTER_FREQUENCY,       
                                             float, 
                                             help = """
                                                    The center frequency of the SDR in Hz.
                                                    This value determines the midpoint of the frequency range
                                                    being processed.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_positive
                                                 ]),
    PNames.MIN_FREQUENCY:          PTemplate(PNames.MIN_FREQUENCY,          
                                             float, 
                                             help = """
                                                    The minimum center frequency, in Hz, for the frequency sweep.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_positive
                                                 ]),
    PNames.MAX_FREQUENCY:          PTemplate(PNames.MAX_FREQUENCY,          
                                             float, 
                                             help = """
                                                    The maximum center frequency, in Hz, for the frequency sweep.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_positive
                                                 ]),
    PNames.FREQUENCY_STEP:         PTemplate(PNames.FREQUENCY_STEP,         
                                             float, 
                                             help = """
                                                    The amount, in Hz, by which the center frequency is incremented
                                                    for each step in the frequency sweep. 
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_positive
                                                 ]),
    PNames.BANDWIDTH:              PTemplate(PNames.BANDWIDTH,              
                                             float, 
                                             help = """
                                                    The frequency range in Hz the signal will occupy without
                                                    significant attenutation.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_positive
                                                 ]),
    PNames.SAMPLE_RATE:            PTemplate(PNames.SAMPLE_RATE,            
                                             int,   
                                             help = """
                                                    The number of samples per second in Hz.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_positive
                                                 ]),
    PNames.IF_GAIN:                PTemplate(PNames.IF_GAIN,                
                                             float, 
                                             help = """
                                                    The intermediate frequency gain, in dB.
                                                    Negative value indicates attenuation.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_negative
                                                 ]),
    PNames.RF_GAIN:                PTemplate(PNames.RF_GAIN,                
                                             float, 
                                             help = """
                                                    The radio frequency gain, in dB.
                                                    Negative value indicates attenuation.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_non_positive
                                                 ]),
    PNames.EVENT_HANDLER_KEY:      PTemplate(PNames.EVENT_HANDLER_KEY,      
                                             str,
                                             help = """
                                                    Identifies which post-processing functions to invoke
                                                    on newly created files.
                                                    """),
    PNames.CHUNK_KEY:              PTemplate(PNames.CHUNK_KEY,              
                                             str,
                                             help = """
                                                    Identifies the type of data is stored in each chunk.
                                                    """,
                                             ),
    PNames.WINDOW_SIZE:            PTemplate(PNames.WINDOW_SIZE,            
                                             int,  
                                             help = """
                                                    The size of the window, in samples, when performing the
                                                    Short Time FFT.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_positive, 
                                                 PConstraints.power_of_two
                                                 ]),
    PNames.WINDOW_HOP:             PTemplate(PNames.WINDOW_HOP,             
                                             int,   
                                             help = """
                                                    How much the window is shifted, in samples, 
                                                    when performing the Short Time FFT.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_positive
                                                 ]),
    PNames.WINDOW_TYPE:            PTemplate(PNames.WINDOW_TYPE,            
                                             str,
                                             help = """
                                                    The type of window applied when performing the Short
                                                    Time FFT.
                                                    """,
                                             ),
    PNames.WATCH_EXTENSION:        PTemplate(PNames.WATCH_EXTENSION,        
                                             str,
                                             help = """
                                                    Post-processing is triggered by newly created files with this extension.
                                                    Extensions are specified without the '.' character.
                                                    """,
                                             ),
    PNames.TIME_RESOLUTION:        PTemplate(PNames.TIME_RESOLUTION,        
                                             float, 
                                             nullable=True,
                                             help = """
                                                    Batched spectrograms are smoothed by averaging up to the time resolution,
                                                    specified in seconds.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_non_negative
                                                 ]),
    PNames.FREQUENCY_RESOLUTION:   PTemplate(PNames.FREQUENCY_RESOLUTION,   
                                             float, 
                                             nullable=True,
                                             help = """
                                                    Batched spectrograms are smoothed by averaging up to the frequency resolution,
                                                    specified in Hz.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_non_negative
                                                 ]),
    PNames.TIME_RANGE:             PTemplate(PNames.TIME_RANGE,             
                                             float, 
                                             nullable=True,
                                             help = """
                                                    Batched spectrograms are stitched together until
                                                    the time range, in seconds, is surpassed.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_non_negative
                                                 ]),
    PNames.BATCH_SIZE:             PTemplate(PNames.BATCH_SIZE,             
                                             int,   
                                             help = """
                                                    SDR data is collected in batches of this size, specified
                                                    in seconds.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_positive
                                                 ]),
    PNames.SAMPLES_PER_STEP:       PTemplate(PNames.SAMPLES_PER_STEP,       
                                             int,   
                                             help = """
                                                    The number of samples taken at each center frequency in the sweep.
                                                    This may vary slightly from what is specified due to the nature of
                                                    GNU Radio runtime.
                                                    """,
                                             pconstraints=[
                                                 PConstraints.enforce_positive
                                                 ]),
    PNames.ORIGIN:                 PTemplate(PNames.ORIGIN,
                                             str,
                                             nullable=True,
                                             help="""
                                                  Corresponds to the FITS keyword ORIGIN.
                                                  """),
    PNames.TELESCOPE:              PTemplate(PNames.TELESCOPE,
                                             str,
                                             nullable=True,
                                             help="""
                                                  Corresponds to the FITS keyword TELESCOP.
                                                  """),
    PNames.INSTRUMENT:             PTemplate(PNames.INSTRUMENT,
                                             str,
                                             nullable=True,
                                             help="""
                                                  Corresponds to the FITS keyword INSTRUME.
                                                  """),
    PNames.OBJECT:                 PTemplate(PNames.OBJECT,
                                             str,
                                             nullable=True,
                                             help="""
                                                  Corresponds to the FITS keyword OBJECT.
                                                  """),
    PNames.OBS_LAT:                PTemplate(PNames.OBS_LAT,
                                             float,
                                             nullable=True,
                                             help="""
                                                  Corresponds to the FITS keyword OBS_LAT.
                                                  """),
    PNames.OBS_LON:                PTemplate(PNames.OBS_LON,
                                             float,
                                             nullable=True,
                                             help="""
                                                  Corresponds to the FITS keyword OBS_LONG.
                                                  """),
    PNames.OBS_ALT:                PTemplate(PNames.OBS_ALT,
                                             float,
                                             nullable=True,
                                             help="""
                                                  Corresponds to the FITS keyword OBS_ALT.
                                                  """),
    PNames.AMPLITUDE:              PTemplate(PNames.AMPLITUDE,
                                             float,
                                             help="""
                                                  The amplitude of the signal.
                                                  """),
    PNames.FREQUENCY:              PTemplate(PNames.FREQUENCY,
                                             float,
                                             help="""
                                                  The frequency of the signal, in Hz.
                                                  """),
    PNames.MIN_SAMPLES_PER_STEP:   PTemplate(PNames.MIN_SAMPLES_PER_STEP,
                                             int,
                                             help="""
                                                  The number of samples in the shortest step of the staircase.
                                                  """,
                                             pconstraints=[
                                                 PConstraints.enforce_positive
                                                 ]),
    PNames.MAX_SAMPLES_PER_STEP:   PTemplate(PNames.MAX_SAMPLES_PER_STEP,
                                             int,
                                             help="""
                                                  The number of samples in the longest step of the staircase.
                                                  """,
                                            pconstraints=[
                                                  PConstraints.enforce_positive
                                                  ]),
    PNames.STEP_INCREMENT:         PTemplate(PNames.STEP_INCREMENT,
                                             int,
                                             help="""
                                                  The length by which each step in the staircase is incremented.
                                                  """,
                                             pconstraints=[
                                                  PConstraints.enforce_positive,
                                              ])
                                    
                                           
}

T = TypeVar('T')
def get_base_ptemplate(
    parameter_name: str,
) -> PTemplate:
    """Create a fresh deep copy of a pre-defined ptemplate"""
    if parameter_name not in _base_ptemplates:
        raise KeyError(f"No ptemplate found for the parameter name '{parameter_name}'. "
                       f"Expected one of {list(_base_ptemplates.keys())}")
    return deepcopy( _base_ptemplates[parameter_name] )


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

import numpy as np
from typing import Tuple

import os

from spectre_core.capture_configs import CaptureConfig, PNames, CaptureModes
from spectre_core.chunks import BaseChunk
from spectre_core.spectrograms import Spectrogram, time_average, frequency_average
from .._base import BaseEventHandler, make_sft_instance
from .._register import register_event_handler


def _do_stfft(iq_data: np.array,
              capture_config: CaptureConfig,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """For reference: https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.ShortTimeFFT.html"""

    sft = make_sft_instance(capture_config)

    # set p0=0, since by convention in the STFFT docs, p=0 corresponds to the slice centred at t=0
    p0=0

    # set p1 to the index of the first slice where the "midpoint" of the window is still inside the signal
    num_samples = len(iq_data)
    p1 = sft.upper_border_begin(num_samples)[1]
    
    # compute a ShortTimeFFT on the IQ samples
    complex_spectra = sft.stft(iq_data, 
                               p0 = p0, 
                               p1 = p1) 
    
    # compute the magnitude of each spectral component
    dynamic_spectra = np.abs(complex_spectra)


    # assign a physical time to each spectrum
    # p0 is defined to correspond with the first sample, at t=0 [s]
    times = sft.t(num_samples, 
                  p0 = p0, 
                  p1 = p1)
    # assign physical frequencies to each spectral component
    frequencies = sft.f + capture_config.get_parameter_value(PNames.CENTER_FREQUENCY) 

    return times, frequencies, dynamic_spectra


def _build_spectrogram(chunk: BaseChunk,
                       capture_config: CaptureConfig) -> Spectrogram:
    """Create a spectrogram by performing a Short Time FFT on the IQ samples for this chunk."""

    # read the data from the chunk
    millisecond_correction = chunk.read_file("hdr")
    iq_data = chunk.read_file("bin")

    # units conversion
    microsecond_correction = millisecond_correction * 1e3

    times, frequencies, dynamic_spectra = _do_stfft(iq_data,
                                                    capture_config)

    # explicitly type cast data arrays to 32-bit floats
    times = np.array(times, dtype = 'float32')
    frequencies = np.array(frequencies, dtype = 'float32')
    dynamic_spectra = np.array(dynamic_spectra, dtype = 'float32')

    return Spectrogram(dynamic_spectra, 
                       times, 
                       frequencies, 
                       chunk.tag, 
                       chunk_start_time = chunk.chunk_start_time, 
                       microsecond_correction = microsecond_correction,
                       spectrum_type = "amplitude")


@register_event_handler(CaptureModes.FIXED_CENTER_FREQUENCY)
class _EventHandler(BaseEventHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def process(self, 
                absolute_file_path: str):
        _LOGGER.info(f"Processing: {absolute_file_path}")
        file_name = os.path.basename(absolute_file_path)
        base_file_name, _ = os.path.splitext(file_name)
        chunk_start_time, tag = base_file_name.split('_')

        # create an instance of the current chunk being processed
        chunk = self._Chunk(chunk_start_time, tag)

        _LOGGER.info("Creating spectrogram")
        spectrogram = _build_spectrogram(chunk,
                                        self._capture_config)

        spectrogram = time_average(spectrogram,
                                   resolution = self._capture_config.get_parameter_value(PNames.TIME_RESOLUTION))

        spectrogram = frequency_average(spectrogram,
                                        resolution = self._capture_config.get_parameter_value(PNames.FREQUENCY_RESOLUTION))
        
        self._cache_spectrogram(spectrogram)

        bin_chunk = chunk.get_file('bin')
        _LOGGER.info(f"Deleting {bin_chunk.file_path}")
        bin_chunk.delete()

        hdr_chunk = chunk.get_file('hdr')
        _LOGGER.info(f"Deleting {hdr_chunk.file_path}")
        hdr_chunk.delete()



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

# Global dictionaries to hold the mappings
event_handler_map = {}

# classes decorated with @register_event_handler([EVENT_HANDLER_KEY])
# will be added to event_handler_map
def register_event_handler(event_handler_key: str):
    def decorator(cls):
        event_handler_map[event_handler_key] = cls
        return cls
    return decorator




# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod
from typing import Callable, Optional
from numbers import Number

from spectre_core.exceptions import ModeNotFoundError
from ._spec_names import SpecNames
from spectre_core.capture_configs import (
    CaptureTemplate, CaptureModes, Parameters, Bound, PValidators, PNames,
    get_base_capture_template, get_base_ptemplate, OneOf, CaptureConfig
)

class BaseReceiver(ABC):
    def __init__(self, 
                 name: str, 
                 mode: Optional[str] = None):
        self._name = name

        self._specs: dict[str, Number | list[Number]] = {}
        self._add_specs()
        
        self._capture_methods: dict[str, Callable] = {}
        self._add_capture_methods()
    
        self._pvalidators: dict[str, Callable] = {}
        self._add_pvalidators()

        self._capture_templates: dict[str, CaptureTemplate] = {}
        self._add_capture_templates()

        self.mode = mode


    @abstractmethod
    def _add_specs(self) -> None:
        pass
    

    @abstractmethod
    def _add_capture_methods(self) -> None:
        pass


    @abstractmethod
    def _add_pvalidators(self) -> None:
        pass


    @abstractmethod
    def _add_capture_templates(self) -> None:
        pass

    
    @property
    def name(self) -> str:
        return self._name
    

    @property
    def capture_methods(self) -> dict[str, Callable]:
        return self._capture_methods
    

    @property
    def pvalidators(self) -> dict[str, Callable]:
        return self._pvalidators


    @property
    def capture_templates(self) -> dict[str, CaptureTemplate]:
        return self._capture_templates


    @property
    def specs(self) -> dict[str, Number]:
        return self._specs


    @property
    def modes(self) -> list[str]:
        capture_method_modes    = list(self.capture_methods.keys())
        pvalidator_modes        = list(self.pvalidators.keys())
        capture_template_modes  = list(self.capture_templates.keys())

        if capture_method_modes == pvalidator_modes == capture_template_modes:
            return capture_method_modes
        else:
            raise ValueError(f"Mode mismatch for the receiver {self.name}.")


    @property
    def mode(self) -> str:
        return self._mode
    

    @mode.setter
    def mode(self, value: Optional[str]) -> None:
        if (value is not None) and value not in self.modes:
            raise ModeNotFoundError((f"{value} is not a defined mode for the receiver {self.name}. "
                                     f"Expected one of {self.modes}"))
        self._mode = value


    @property
    def capture_method(self) -> Callable:
        return self.capture_methods[self.mode]


    @property
    def pvalidator(self) -> Callable:
        return self.pvalidators[self.mode]


    @property
    def capture_template(self) -> CaptureTemplate:
        return self._capture_templates[self.mode]


    def add_capture_method(self,
                           mode: str,
                           capture_method: Callable) -> None:
        self._capture_methods[mode] = capture_method


    def add_pvalidator(self,
                      mode: str,
                      pvalidator: Callable) -> None:
        self._pvalidators[mode] = pvalidator


    def add_capture_template(self,
                             mode: str,
                             capture_template: CaptureTemplate) -> CaptureTemplate:
        self._capture_templates[mode] = capture_template


    def add_spec(self,
                 name: str,
                 value: Number) -> None:
        self.specs[name] = value


    def get_spec(self, 
                 spec_name: str) -> Number | list[Number]:
        if spec_name not in self.specs:
            raise KeyError(f"Spec not found with name '{spec_name}' "
                           f"for the receiver '{self.name}'")
        return self.specs[spec_name]


    def start_capture(self, 
                      tag: str) -> None:
        self.capture_method( tag, self.load_parameters(tag) )


    def save_parameters(self,
                        tag: str,
                        parameters: Parameters,
                        force: bool = False) -> None:
        
        parameters = self.capture_template.apply_template(parameters)
        self.pvalidator(parameters)

        capture_config = CaptureConfig(tag)
        capture_config.save_parameters(self.name,
                                       self.mode,
                                       parameters,
                                       force=force)

    def load_parameters(self,
                        tag: str) -> Parameters:
        capture_config = CaptureConfig(tag)

        parameters = self.capture_template.apply_template(capture_config.parameters)
        self.pvalidator(parameters)
        
        return parameters


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Optional
from dataclasses import dataclass
from functools import partial

from spectre_core.capture_configs import (
    CaptureModes
)
from ..gr._rspduo import CaptureMethods
from .._spec_names import SpecNames
from ._sdrplay_receiver import SDRPlayReceiver
from .._register import register_receiver

@dataclass
class Modes:
    TUNER_1_FIXED_CENTER_FREQUENCY  = f"tuner-1-{CaptureModes.FIXED_CENTER_FREQUENCY}"
    TUNER_2_FIXED_CENTER_FREQUENCY  = f"tuner-2-{CaptureModes.FIXED_CENTER_FREQUENCY}"
    TUNER_1_SWEPT_CENTER_FREQUENCY  = f"tuner-1-{CaptureModes.SWEPT_CENTER_FREQUENCY}"


@register_receiver("rspduo")
class _Receiver(SDRPlayReceiver):
    def __init__(self, 
                 name: str,
                 mode: Optional[str]):
        super().__init__(name,
                         mode)
        

    def _add_specs(self) -> None:
        self.add_spec( SpecNames.SAMPLE_RATE_LOWER_BOUND, 200e3     ) # Hz
        self.add_spec( SpecNames.SAMPLE_RATE_UPPER_BOUND, 10e6      ) # Hz
        self.add_spec( SpecNames.FREQUENCY_LOWER_BOUND  , 1e3       ) # Hz
        self.add_spec( SpecNames.FREQUENCY_UPPER_BOUND  , 2e9       ) # Hz
        self.add_spec( SpecNames.IF_GAIN_UPPER_BOUND    , -20       ) # dB
        self.add_spec( SpecNames.RF_GAIN_UPPER_BOUND    , 0         ) # dB
        self.add_spec( SpecNames.API_RETUNING_LATENCY   , 50 * 1e-3 ) # s
        self.add_spec( SpecNames.BANDWIDTH_OPTIONS, 
                      [200000, 300000, 600000, 1536000, 5000000, 6000000, 7000000, 8000000])


    def _add_capture_methods(self) -> None:
        self.add_capture_method(Modes.TUNER_1_FIXED_CENTER_FREQUENCY, 
                                CaptureMethods.tuner_1_fixed_center_frequency)
        self.add_capture_method(Modes.TUNER_2_FIXED_CENTER_FREQUENCY, 
                                CaptureMethods.tuner_2_fixed_center_frequency)
        self.add_capture_method(Modes.TUNER_1_SWEPT_CENTER_FREQUENCY, 
                                CaptureMethods.tuner_1_swept_center_frequency)
    

    def _add_capture_templates(self):
        self.add_capture_template(Modes.TUNER_1_FIXED_CENTER_FREQUENCY,
                                  self._get_capture_template_fixed_center_frequency())
        self.add_capture_template(Modes.TUNER_2_FIXED_CENTER_FREQUENCY,
                                  self._get_capture_template_fixed_center_frequency())
        self.add_capture_template(Modes.TUNER_1_SWEPT_CENTER_FREQUENCY,
                                  self._get_capture_template_swept_center_frequency())
        
    
    def _add_pvalidators(self):
        self.add_pvalidator(Modes.TUNER_1_FIXED_CENTER_FREQUENCY,
                            self._get_pvalidator_fixed_center_frequency())
        self.add_pvalidator(Modes.TUNER_2_FIXED_CENTER_FREQUENCY,
                            self._get_pvalidator_fixed_center_frequency())
        self.add_pvalidator(Modes.TUNER_1_SWEPT_CENTER_FREQUENCY,
                            self._get_pvalidator_swept_center_frequency())



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from typing import Optional, Callable

from spectre_core.capture_configs import (
    CaptureTemplate, CaptureModes, Parameters, Bound, PValidators, PNames,
    get_base_capture_template, make_base_capture_template, get_base_ptemplate
)
from ..gr._test import CaptureMethods
from .._spec_names import SpecNames
from .._base import BaseReceiver
from .._register import register_receiver


@dataclass
class Modes:
    COSINE_SIGNAL_1  = "cosine-signal-1"
    TAGGED_STAIRCASE = "tagged-staircase"


@register_receiver("test")
class _Receiver(BaseReceiver):
    def __init__(self, 
                 name: str,
                 mode: Optional[str]):
        super().__init__(name,
                         mode)
        

    def _add_specs(self) -> None:
        self.add_spec( SpecNames.SAMPLE_RATE_LOWER_BOUND, 64000  )
        self.add_spec( SpecNames.SAMPLE_RATE_UPPER_BOUND, 640000 )
        self.add_spec( SpecNames.FREQUENCY_LOWER_BOUND  , 16000  )
        self.add_spec( SpecNames.FREQUENCY_UPPER_BOUND  , 160000 )


    def _add_capture_methods(self) -> None:
        self.add_capture_method( Modes.COSINE_SIGNAL_1 , CaptureMethods.cosine_signal_1 )
        self.add_capture_method( Modes.TAGGED_STAIRCASE, CaptureMethods.tagged_staircase)
    

    def _add_pvalidators(self) -> None:
        self.add_pvalidator( Modes.COSINE_SIGNAL_1 , self.__get_pvalidator_cosine_signal_1()  )
        self.add_pvalidator( Modes.TAGGED_STAIRCASE, self.__get_pvalidator_tagged_staircase() )


    def _add_capture_templates(self) -> None:
        self.add_capture_template( Modes.COSINE_SIGNAL_1 , self.__get_capture_template_cosine_signal_1()  )
        self.add_capture_template( Modes.TAGGED_STAIRCASE, self.__get_capture_template_tagged_staircase() )
        

    def __get_capture_template_cosine_signal_1(self) -> CaptureTemplate:
        #
        # Create the base template
        #
        capture_template = get_base_capture_template( CaptureModes.FIXED_CENTER_FREQUENCY )
        capture_template.add_ptemplate( get_base_ptemplate(PNames.AMPLITUDE) )
        capture_template.add_ptemplate( get_base_ptemplate(PNames.FREQUENCY) )

        #
        # Update the defaults
        #
        capture_template.set_defaults(
            (PNames.BATCH_SIZE,            3.0),
            (PNames.CENTER_FREQUENCY,      16000),
            (PNames.AMPLITUDE,             2.0),
            (PNames.FREQUENCY,             32000),
            (PNames.SAMPLE_RATE,           128000),
            (PNames.WINDOW_HOP,            512),
            (PNames.WINDOW_SIZE,           512),
            (PNames.WINDOW_TYPE,           "boxcar")
        )

        #
        # Enforce defaults
        #
        capture_template.enforce_defaults(
            PNames.TIME_RESOLUTION,
            PNames.TIME_RANGE,
            PNames.FREQUENCY_RESOLUTION,
            PNames.WINDOW_TYPE
        )


        #
        # Adding pconstraints
        #
        capture_template.add_pconstraint(
            PNames.SAMPLE_RATE,
            [
                Bound(
                    lower_bound=self.get_spec(SpecNames.SAMPLE_RATE_LOWER_BOUND),
                    upper_bound=self.get_spec(SpecNames.SAMPLE_RATE_UPPER_BOUND)
                )
            ]
        )
        capture_template.add_pconstraint(
            PNames.FREQUENCY,
            [
                Bound(
                    lower_bound=self.get_spec(SpecNames.FREQUENCY_LOWER_BOUND),
                    upper_bound=self.get_spec(SpecNames.FREQUENCY_UPPER_BOUND)
                )
            ]
        )
        return capture_template


    def __get_capture_template_tagged_staircase(self) -> CaptureTemplate:
        #
        # Make the base template
        #
        capture_template = make_base_capture_template(
                PNames.TIME_RESOLUTION,
                PNames.FREQUENCY_RESOLUTION,
                PNames.TIME_RANGE,
                PNames.SAMPLE_RATE,
                PNames.BATCH_SIZE,
                PNames.WINDOW_TYPE,
                PNames.WINDOW_HOP,
                PNames.WINDOW_SIZE,
                PNames.EVENT_HANDLER_KEY,
                PNames.CHUNK_KEY,
                PNames.WATCH_EXTENSION,
                PNames.MIN_SAMPLES_PER_STEP,
                PNames.MAX_SAMPLES_PER_STEP,
                PNames.FREQUENCY_STEP,
                PNames.STEP_INCREMENT,
                PNames.OBS_ALT,
                PNames.OBS_LAT,
                PNames.OBS_LON,
                PNames.OBJECT,
                PNames.ORIGIN,
                PNames.TELESCOPE,
                PNames.INSTRUMENT
        )

        #
        # Update the defaults
        #
        capture_template.set_defaults(
            (PNames.BATCH_SIZE,            3.0),
            (PNames.FREQUENCY_STEP,        128000),
            (PNames.MAX_SAMPLES_PER_STEP,  5000),
            (PNames.MIN_SAMPLES_PER_STEP,  4000),
            (PNames.SAMPLE_RATE,           128000),
            (PNames.STEP_INCREMENT,        200),
            (PNames.WINDOW_HOP,            512),
            (PNames.WINDOW_SIZE,           512),
            (PNames.WINDOW_TYPE,           "boxcar"),
            (PNames.EVENT_HANDLER_KEY,     CaptureModes.SWEPT_CENTER_FREQUENCY),
            (PNames.CHUNK_KEY,             CaptureModes.SWEPT_CENTER_FREQUENCY),
            (PNames.WATCH_EXTENSION,       "bin")
        )


        #
        # Enforce defaults
        #
        capture_template.enforce_defaults(
            PNames.TIME_RESOLUTION,
            PNames.TIME_RANGE,
            PNames.FREQUENCY_RESOLUTION,
            PNames.WINDOW_TYPE,
            PNames.EVENT_HANDLER_KEY,
            PNames.CHUNK_KEY,
            PNames.WATCH_EXTENSION
        )

        return capture_template
    
    
    def __get_pvalidator_cosine_signal_1(self) -> Callable:
        def pvalidator(parameters: Parameters):
            PValidators.window(parameters)

            sample_rate          = parameters.get_parameter_value(PNames.SAMPLE_RATE)
            frequency            = parameters.get_parameter_value(PNames.FREQUENCY)
            window_size          = parameters.get_parameter_value(PNames.WINDOW_SIZE)

            # check that the sample rate is an integer multiple of the underlying signal frequency
            if sample_rate % frequency != 0:
                raise ValueError("The sampling rate must be some integer multiple of frequency")

            a = sample_rate/frequency
            if a < 2:
                raise ValueError((f"The ratio of sampling rate over frequency must be greater than two. "
                                f"Got {a}"))
            
            # analytical requirement
            # if p is the number of sampled cycles, we can find that p = window_size / a
            # the number of sampled cycles must be a positive natural number.
            p = window_size / a
            if window_size % a != 0:
                raise ValueError((f"The number of sampled cycles must be a positive natural number. "
                                  f"Computed that p={p}"))
        return pvalidator


    def __get_pvalidator_tagged_staircase(self) -> None:
        def pvalidator(parameters: Parameters):
            PValidators.window(parameters)

            freq_step            = parameters.get_parameter_value(PNames.FREQUENCY_STEP)
            sample_rate          = parameters.get_parameter_value(PNames.SAMPLE_RATE)
            min_samples_per_step = parameters.get_parameter_value(PNames.MIN_SAMPLES_PER_STEP)
            max_samples_per_step = parameters.get_parameter_value(PNames.MAX_SAMPLES_PER_STEP)

            if freq_step != sample_rate:
                raise ValueError(f"The frequency step must be equal to the sampling rate")
            
            
            if min_samples_per_step > max_samples_per_step:
                raise ValueError((f"Minimum samples per step cannot be greater than the maximum samples per step. "
                                f"Got {min_samples_per_step}, which is greater than {max_samples_per_step}"))
            
        return pvalidator
        


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime

import numpy as np

def average_array(array: np.ndarray, average_over: int, axis=0) -> np.ndarray:

    # Check if average_over is an integer
    if type(average_over) != int:
        raise TypeError(f"average_over must be an integer. Got {type(average_over)}")

    # Get the size of the specified axis which we will average over
    axis_size = array.shape[axis]
    # Check if average_over is within the valid range
    if not 1 <= average_over <= axis_size:
        raise ValueError(f"average_over must be between 1 and the length of the axis ({axis_size})")
    
    max_axis_index = len(np.shape(array)) - 1
    if axis > max_axis_index: # zero indexing on specifying axis, so minus one
        raise ValueError(f"Requested axis is out of range of array dimensions. Axis: {axis}, max axis index: {max_axis_index}")

    # find the number of elements in the requested axis
    num_elements = array.shape[axis]

    # find the number of "full blocks" to average over
    num_full_blocks = num_elements // average_over
    # if num_elements is not exactly divisible by average_over, we will have some elements left over
    # these remaining elements will be padded with nans to become another full block
    remainder = num_elements % average_over
    
    # if there exists a remainder, pad the last block
    if remainder != 0:
        # initialise an array to hold the padding shape
        padding_shape = [(0, 0)] * array.ndim
        # pad after the last column in the requested axis
        padding_shape[axis] = (0, average_over - remainder)
        # pad with nan values (so to not contribute towards the mean computation)
        array = np.pad(array, padding_shape, mode='constant', constant_values=np.nan)
    
    # initalise a list to hold the new shape
    new_shape = list(array.shape)
    # update the shape on the requested access (to the number of blocks we will average over)
    new_shape[axis] = num_full_blocks + (1 if remainder else 0)
    # insert a new dimension, with the size of each block
    new_shape.insert(axis + 1, average_over)
    # and reshape the array to sort the array into the relevant blocks.
    reshaped_array = array.reshape(new_shape)
    # average over the newly created axis, essentially averaging over the blocks.
    averaged_array = np.nanmean(reshaped_array, axis=axis + 1)
    # return the averaged array
    return averaged_array


def is_close(ar: np.ndarray, 
             ar_comparison: np.ndarray,
             absolute_tolerance: float) -> bool:
    """Close enough accounts for wiggle-room equating floats."""
    return np.all(np.isclose(ar, 
                             ar_comparison, 
                             atol=absolute_tolerance))

def find_closest_index(
    target_value: float | datetime, 
    array: np.ndarray, 
    enforce_strict_bounds: bool = False
) -> int:
    # Ensure input array is a numpy array
    array = np.asarray(array)

    # Convert to datetime64 if necessary
    if isinstance(target_value, datetime) or np.issubdtype(array.dtype, np.datetime64):
        target_value = np.datetime64(target_value)
        array = array.astype('datetime64[ns]')
    else:
        target_value = float(target_value)
        array = array.astype(float)

    # Check bounds if strict enforcement is required
    if enforce_strict_bounds:
        max_value, min_value = np.nanmax(array), np.nanmin(array)
        if target_value > max_value:
            raise ValueError(f"Target value {target_value} exceeds max array value {max_value}")
        if target_value < min_value:
            raise ValueError(f"Target value {target_value} is less than min array value {min_value}")

    # Find the index of the closest value
    return np.argmin(np.abs(array - target_value))


def normalise_peak_intensity(array: np.ndarray) -> np.ndarray:
    return array/np.nanmax(array)


def compute_resolution(array: np.ndarray) -> float:
    # Check that the array is one-dimensional
    if array.ndim != 1:
        raise ValueError("Input array must be one-dimensional")
    
    if len(array) < 2:
        raise ValueError("Input array must contain at least two elements")
    
    # Calculate differences between consecutive elements.
    resolutions = np.diff(array)

    return np.nanmedian(resolutions)


def compute_range(array: np.ndarray) -> float:
    # Check that the array is one-dimensional
    if array.ndim != 1:
        raise ValueError("Input array must be one-dimensional")
    
    if len(array) < 2:
        raise ValueError("Input array must contain at least two elements")
    return array[-1] - array[0]


def subtract_background(array: np.ndarray, start_index: int, end_index: int) -> np.ndarray:
    array -= np.nanmean(array[start_index:end_index+1])
    return array


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# SPDX-License-Identifier: GPL-3.0
#
# GNU Radio Python Flow Graph
# Title: Test receiver
# GNU Radio version: 3.10.1.1

# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

#
# Test receiver top blocks
#

from functools import partial
from dataclasses import dataclass

from gnuradio import gr
from gnuradio import blocks
from gnuradio import spectre
from gnuradio import analog

from spectre_core.capture_configs import Parameters, PNames
from spectre_core.config import get_chunks_dir_path
from ._base import capture


class _cosine_signal_1(gr.top_block):
    def __init__(self, 
                 tag: str,
                 parameters: Parameters):
        gr.top_block.__init__(self, catch_exceptions=True)

        ##################################################
        # Unpack capture config
        ##################################################
        samp_rate   = parameters.get_parameter_value(PNames.SAMPLE_RATE)
        batch_size  = parameters.get_parameter_value(PNames.BATCH_SIZE)
        frequency   = parameters.get_parameter_value(PNames.FREQUENCY)
        amplitude   = parameters.get_parameter_value(PNames.AMPLITUDE)

        ##################################################
        # Blocks
        ##################################################
        self.spectre_batched_file_sink_0 = spectre.batched_file_sink(get_chunks_dir_path(), 
                                                                     tag, 
                                                                     batch_size, 
                                                                     samp_rate)
        self.blocks_throttle_0_1 = blocks.throttle(gr.sizeof_float*1, 
                                                   samp_rate,
                                                   True)
        self.blocks_throttle_0 = blocks.throttle(gr.sizeof_float*1, 
                                                 samp_rate,
                                                 True)
        self.blocks_null_source_1 = blocks.null_source(gr.sizeof_float*1)
        self.blocks_float_to_complex_1 = blocks.float_to_complex(1)
        self.analog_sig_source_x_0 = analog.sig_source_f(samp_rate, 
                                                         analog.GR_COS_WAVE, 
                                                         frequency, 
                                                         amplitude, 
                                                         0, 
                                                         0)


        ##################################################
        # Connections
        ##################################################
        self.connect((self.analog_sig_source_x_0, 0), (self.blocks_throttle_0, 0))
        self.connect((self.blocks_float_to_complex_1, 0), (self.spectre_batched_file_sink_0, 0))
        self.connect((self.blocks_null_source_1, 0), (self.blocks_throttle_0_1, 0))
        self.connect((self.blocks_throttle_0, 0), (self.blocks_float_to_complex_1, 0))
        self.connect((self.blocks_throttle_0_1, 0), (self.blocks_float_to_complex_1, 1))


class _tagged_staircase(gr.top_block):
    def __init__(self, 
                 tag: str,
                 parameters: Parameters):
        gr.top_block.__init__(self, catch_exceptions=True)

        ##################################################
        # Unpack capture config
        ##################################################
        step_increment       = parameters.get_parameter_value(PNames.STEP_INCREMENT)
        samp_rate            = parameters.get_parameter_value(PNames.SAMPLE_RATE)
        min_samples_per_step = parameters.get_parameter_value(PNames.MIN_SAMPLES_PER_STEP)
        max_samples_per_step = parameters.get_parameter_value(PNames.MAX_SAMPLES_PER_STEP)
        frequency_step       = parameters.get_parameter_value(PNames.FREQUENCY_STEP)
        batch_size           = parameters.get_parameter_value(PNames.BATCH_SIZE)

        ##################################################
        # Blocks
        ##################################################
        self.spectre_tagged_staircase_0 = spectre.tagged_staircase(min_samples_per_step, 
                                                                   max_samples_per_step, 
                                                                   frequency_step,
                                                                   step_increment, 
                                                                   samp_rate)
        self.spectre_batched_file_sink_0 = spectre.batched_file_sink(get_chunks_dir_path(),
                                                                     tag, 
                                                                     batch_size, 
                                                                     samp_rate, 
                                                                     True,
                                                                     'rx_freq',
                                                                     0) # zero means the center frequency is unset
        self.blocks_throttle_0 = blocks.throttle(gr.sizeof_gr_complex*1, samp_rate, True)


        ##################################################
        # Connections
        ##################################################
        self.connect((self.blocks_throttle_0, 0), (self.spectre_batched_file_sink_0, 0))
        self.connect((self.spectre_tagged_staircase_0, 0), (self.blocks_throttle_0, 0))


@dataclass(frozen=True)
class CaptureMethods:
    cosine_signal_1  = partial(capture, top_block_cls=_cosine_signal_1)
    tagged_staircase = partial(capture, top_block_cls=_tagged_staircase)